"""Paraver related trace generation.

All the Paraver-files (.prv) generation is done through decorators and
mechanisms defined in this module. There are also capabilities for merge and
checks on the generated prv files.

Additionally, this module defines an application capable of performing several
Paraver-related routines, like "merge".
"""
from __future__ import print_function
import os
import types
from distutils.util import strtobool
from functools import wraps
from dataclay.commonruntime.Settings import settings
import logging
import traceback
import importlib
from dataclay.contrib.dummy_pycompss import task
import ctypes
from ctypes import cdll

# Special wrapper of dataclay extrae to initialize it
ExtraeWrapperLib = os.getenv("DATACLAY_EXTRAE_WRAPPER_LIB")

# Explicit and manually crafted list of CLASSES to be instrumentated
CLASSES_WITH_EXTRAE_DECORATORS = {  # Similar to Java paraver/extrae AspectJ file. 
    # MAPPING: package name - modules with class named equal to modulen name
    "dataclay.commonruntime" : ["ClientRuntime", "DataClayRuntime", "ExecutionEnvironmentRuntime" ],
    "dataclay.executionenv" : ["ExecutionEnvironment"],
    "dataclay.heap" : [ "ClientHeapManager", "ExecutionEnvironmentHeapManager", "HeapManager"],
    "dataclay": ["DataClayObject"],
    "dataclay.serialization.lib": ["SerializationLibUtils", "DeserializationLibUtils"],
    "dataclay.serialization.python.util": ["PyTypeWildcardWrapper"],
}

PARAVER_FUNC_MAP = {}
TRACED_METHODS = list()
TRACING_ENABLED = False
LOGGER = logging.getLogger(__name__)
TRACE_ENABLED = strtobool(os.getenv("PARAVER_TRACE_ENABLED", "False"))  # WARNING: not used
TASK_EVENTS = 8000010
CURRENT_AVAILABLE_TASK_ID = 0
TASK_ID = 0
PYEXTRAE = None
EXTRAE_DICT = dict()
DATACLAY_EXTRAE_WRAPPER = None

# Extrae options
EXTRAE_DISABLE_ALL_OPTIONS = 0
EXTRAE_CALLER_OPTION = 1
EXTRAE_HWC_OPTION = 2
EXTRAE_MPI_HWC_OPTION = 4
EXTRAE_MPI_OPTION = 8
EXTRAE_OMP_OPTION = 16
EXTRAE_OMP_HWC_OPTION = 32
EXTRAE_UF_HWC_OPTION = 64
EXTRAE_PTHREAD_OPTION = 128
EXTRAE_PTHREAD_HWC_OPTION = 256
EXTRAE_SAMPLING_OPTION = 512
EXTRAE_ENABLE_ALL_OPTIONS = EXTRAE_CALLER_OPTION | EXTRAE_HWC_OPTION | EXTRAE_MPI_HWC_OPTION | EXTRAE_MPI_OPTION | EXTRAE_OMP_OPTION | EXTRAE_OMP_HWC_OPTION | EXTRAE_UF_HWC_OPTION | EXTRAE_PTHREAD_OPTION | EXTRAE_PTHREAD_HWC_OPTION | EXTRAE_SAMPLING_OPTION


def add_function_trace(enter, prv_value):
    """
    Trace entry: Function (enter or exit).
    :param enter: Boolean to signal whether it is entering or exiting the method.
    :param prv_value: Numerical identifier for PyCOMPSs-compliant Paraver tracing.
    """

    if enter:
        LOGGER.debug("Tracing event with value %i" % (prv_value))
        PYEXTRAE.event(TASK_EVENTS, prv_value)
    else:
        PYEXTRAE.event(TASK_EVENTS, 0)


def add_network_send(trace_time, network_type, send_port, request_id, dest_host_ip, dest_host_port, message_size, method_id):
    """Trace a certain network send (can be a SEND_REQUEST or a SEND_RESPONSE)."""
    if not extrae_tracing_is_enabled():
        return


def add_network_receive(origin_host_ip, origin_host_port, request_id, method_id):
    """Trace a certain network receive (a RESPONSE)."""
    if not extrae_tracing_is_enabled():
        return


def initialize_extrae(initialize=False):
    """ 
    Initializes Extrae tracing. This function is triggered by: 
        starting task id property specified in storage.conf OR
        explicit call from Execution Environment (due to a client activation)
    This function will first check if Extrae is available, if not it will raise an exception. 
    Then, classes specified will be decorated. 
    :param initialize: indicates if library must be initialized. False in case of compss workers.
    :type initialize: boolean
    """
    global TASK_ID
    global CURRENT_AVAILABLE_TASK_ID
    global TRACING_ENABLED
    global CLASSES_WITH_EXTRAE_DECORATORS
    global PYEXTRAE
    global EXTRAE_DICT
    TASK_ID = CURRENT_AVAILABLE_TASK_ID
    LOGGER.debug("Initializing Extrae with task id %i in process with pid %s " % (TASK_ID, str(os.getpid())))

    # This is something very slow which should be done during initialization
    # and not during tracing.
    try:
        import pyextrae.common.extrae as pyextrae_module 
        from pyextrae.common.extrae import Extrae as extrae_dict_obj
        PYEXTRAE = pyextrae_module
        EXTRAE_DICT = extrae_dict_obj

        LOGGER.info("Using pyextrae.common for tracing information")
    except ImportError:
        LOGGER.warning("Trying to activate pyextrae but ImportError happened. Please make sure pyextrae is installed.")
        return 
    
    """ Initialize synchronization events """
    if initialize:
        TracingLibrary = "libseqtrace.so"
        LOGGER.debug("Initializing Extrae -- Adding synchronization event ")
        
        # set task ID 
        num_tasks = TASK_ID + 1
        DATACLAY_EXTRAE_WRAPPER = cdll.LoadLibrary(ExtraeWrapperLib)
        DATACLAY_EXTRAE_WRAPPER.set_task_id(ctypes.c_uint(TASK_ID))
        DATACLAY_EXTRAE_WRAPPER.set_num_tasks(ctypes.c_uint(num_tasks))
        
        # init tracing
        PYEXTRAE.startTracing(TracingLibrary, False)
        EXTRAE_DICT[os.getpid()].Extrae_init()
        
        CURRENT_AVAILABLE_TASK_ID = CURRENT_AVAILABLE_TASK_ID + 1  # only increment in servers

    # initialize type and values 
    load_type_values()
    
    """
    Apply decorators 
    Note that all non-underscore-starting methods are considered public. Take
    that in mind while programming your class. You can also explicitly avoid
    tracing a certain method by doing the following:

        def public_method_not_traced(self, ...):
            ...
        public_method_not_traced.do_not_trace = True
    """
    for module, classes in CLASSES_WITH_EXTRAE_DECORATORS.items():
        for class_name in classes:
            actual_module_name = "%s.%s" % (module, class_name)  # Module name is actually the class name
            try:
                LOGGER.debug("Looking for module %s" % (actual_module_name))
                cur_module = importlib.import_module(actual_module_name)
                klass = getattr(cur_module, class_name)
                LOGGER.debug("Decorating class %r" % (klass))
                for name, obj in klass.__dict__.items():
                    if name.startswith("_"):  # ignore private methods
                        continue
            
                    if isinstance(obj, types.FunctionType):  # @abarcelo doesn't understand why UnboundMethodType wasn't working
                        if getattr(obj, "do_not_trace", False):
                            # The user explicitly defined "do_not_trace" flag here
                            continue
                        LOGGER.debug("Decorating function %r" % (obj))
                        setattr(klass, name, _trace_method_in_class(actual_module_name, class_name, obj))
            except ImportError:
                LOGGER.debug("No module %s found" % actual_module_name)
            except:
                LOGGER.debug("Exception while trying to get class %s from module %s" % (class_name, actual_module_name))
                traceback.print_exc()
    
    LOGGER.debug("Extrae is ready for task %i!" % TASK_ID)

    # Extrae is ready
    TRACING_ENABLED = True

        
def finish_tracing(finalize_extrae=False):
    """ Finishing tracing.
    :param finalize_extrae: indicates extrae must be finalized (False for COMPSs doing it for us)
    :type finalize_extrae: boolean
    """
    global PYEXTRAE
    global EXTRAE_DICT
    global TRACING_ENABLED
    global TASK_ID
    global MERGER_TASK_EVENTS
    global SYNC_EVENTS
    if TRACING_ENABLED:
        LOGGER.debug("Pyextrae pids: %s" % str(PYEXTRAE.Extrae.keys()))
        LOGGER.debug("Finishing tracing with task id %i" % TASK_ID)
        try:
            define_event_types()
            if finalize_extrae:
                # EXTRAE_DICT[os.getpid()].Extrae_set_options(EXTRAE_ENABLE_ALL_OPTIONS & ~EXTRAE_PTHREAD_OPTION)
                PYEXTRAE.pyEx_trace_fini(Master=True)
                # EXTRAE_DICT[os.getpid()].Extrae_set_options(EXTRAE_DISABLE_ALL_OPTIONS)

        except:
            traceback.print_exc()
        LOGGER.debug("Finished tracing with task id %i" % TASK_ID)
    
        TRACING_ENABLED = False


def define_event_types():
    global TRACED_METHODS
    global EXTRAE_DICT
    nvalues = len(TRACED_METHODS) + 1
    LOGGER.debug("Defining event types. Number of traced events: %i" % nvalues)
    description = "dataClay"
    values = (ctypes.c_ulonglong * nvalues)()
    description_values = (ctypes.c_char_p * nvalues)()
    values[0] = 0
    description_values[0] = "End".encode('utf-8')
    for i in range(1, nvalues):
        try:
            description_values[i] = TRACED_METHODS[i - 1].encode('utf-8')
            values[i] = PARAVER_FUNC_MAP[description_values[i].decode("utf-8")]
            LOGGER.verbose("Defined event %s with value %s" % (str(description_values[i]), str(values[i])))
        except KeyError:
            LOGGER.info("Tried to load untraced paraver event")

    EXTRAE_DICT[os.getpid()].Extrae_define_event_type(ctypes.pointer(ctypes.c_uint(TASK_EVENTS)),
                                       ctypes.c_char_p(description.encode('utf-8')),
                                       ctypes.pointer(ctypes.c_uint(nvalues)),
                                       ctypes.pointer(values),
                                       ctypes.pointer(description_values))


def enable_pthreads():
    """ Enables extrae pthreads"""
    EXTRAE_DICT[os.getpid()].Extrae_set_options(EXTRAE_DISABLE_ALL_OPTIONS)

    
def disable_pthreads():
    """ Disable extrae pthreads"""
    EXTRAE_DICT[os.getpid()].Extrae_set_options(EXTRAE_ENABLE_ALL_OPTIONS & ~EXTRAE_PTHREAD_OPTION)


def enable_extrae_tracing():
    """ Enables extrae tracing. """ 
    global TRACING_ENABLED
    TRACING_ENABLED = True


def get_traces():
    """ At this point, pyextrae will call merger to create PRV file. 
    This PRV file must be send to client/master in order to be merged with other python EE prvs and
    Java services. This function is not called for python COMPSs workers.
    :returns: prv files of current host 
    :rtype: dict<string, bytes> 
    """
    global TASK_EVENTS
    global PARAVER_FUNC_MAP
    global TRACED_METHODS
    traces = dict()  
    global TASK_ID
    set_path = Configuration.TRACES_DEST_PATH + "/set-0"
    LOGGER.debug("Sending files in %s" % set_path)
    for dirpath, subdirs, files in os.walk(set_path):
        for name in files:
            mpitfile = os.path.join(dirpath, name)
            LOGGER.debug("Found file %s" % name)
            in_file = open(mpitfile, "rb")  # opening for [r]eading as [b]inary
            data = in_file.read() 
            in_file.close()
            traces[name] = data

    return traces

        
def extrae_tracing_is_enabled():
    """
    Indicates if Extrae tracing is enabled. 
    :return: True if extrae tracing is enabled, false otherwise. 
    :rtype: boolean
    """ 
    global TRACING_ENABLED
    return TRACING_ENABLED

        
def set_current_available_task_id(task_id):
    """
    Sets current available task id 
    :param task_id: The current available task id 
    :type task_id: integer
    """
    global CURRENT_AVAILABLE_TASK_ID
    CURRENT_AVAILABLE_TASK_ID = task_id

        
def get_current_available_task_id():
    """
    Get current available task ID 
    :return: Current available task ID 
    :rtype: integer
    """
    global CURRENT_AVAILABLE_TASK_ID
    return CURRENT_AVAILABLE_TASK_ID


def get_task_id():
    """
    Get  task ID 
    :return: The task ID 
    :rtype: integer
    """
    global TASK_ID
    return TASK_ID


def _is_network_tracing_enabled():
    """Check if network calls tracing is enabled (globally for all calls).

    :return: True if it should be traced, false otherwise.
    """
    global TRACE_ENABLED
    return TRACE_ENABLED


def _trace_method_in_class(module_name, class_name, func):
    """Decorator for methods in a class, when class is known."""
    global TRACED_METHODS
    global PARAVER_FUNC_MAP
    try:
        descriptor = "%s.%s.%s" % (module_name, class_name, func.__name__)
        LOGGER.debug("Looking for descriptor %s in paraver intercepted methods" % descriptor)
        prv_value = PARAVER_FUNC_MAP[descriptor]
        LOGGER.debug("Found paraver value %r for descriptor %s in paraver intercepted methods" % (prv_value, descriptor))
    except KeyError:
        LOGGER.warning("Method `%s` (class %s, module %s) "
                       "is not correctly registered for Paraver",
                       func.__name__, class_name, module_name)
        prv_value = 999

    @wraps(func)
    def func_wrapper(self, *args, **kwargs):
        
        if extrae_tracing_is_enabled():
            TRACED_METHODS.append(descriptor)
            add_function_trace(True, prv_value)
        # Proceed to the method call
        try:
            ret = func(self, *args, **kwargs)
        finally:
            if extrae_tracing_is_enabled():
                add_function_trace(False, prv_value)
        return ret
    
    return func_wrapper


def load_type_values():
    global PARAVER_FUNC_MAP

    __location__ = os.path.realpath(
        os.path.join(Configuration.TRACES_DEST_PATH, os.path.dirname(__file__)))
    prv_values_file = os.path.join(__location__, 'python_paraver_values.properties')
    LOGGER.debug("Loading paraver values file from %s" % str(prv_values_file))
    with open(prv_values_file, 'r') as f:
        for line in f:
            line = line.rstrip()  # removes trailing whitespace and '\n' chars
            if "=" not in line: continue  # skips blanks and comments w/o =
            if line.startswith("#"): continue  # skips comments which contain =
            k, v = line.split("=", 1)
            PARAVER_FUNC_MAP[k] = int(v)
    LOGGER.debug("Loaded %i functions to intercept" % len(PARAVER_FUNC_MAP))



from __future__ import absolute_import
""" Class description goes here. """

from dataclay import StorageObject
from dataclay import dclayMethod

from .splitting import SplittableCollectionMixin

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2017 Barcelona Supercomputing Center (BSC-CNS)'

CLASSES_TO_REGISTER = (
    "StorageList", "ListChunk", "StorageDict", "DictChunk"
)


class ListChunk(StorageObject):
    """
    @ClassField items list<storageobject>
    """

    @dclayMethod(elements="list")
    def __init__(self, elements):
        self.items = elements[:]

    @dclayMethod(idx=int, return_="anything")
    def __getitem__(self, idx):
        return self.items[idx]

    @dclayMethod(return_="anything", _local=True)
    def __iter__(self):
        # Note that the following cannot be serialized, hence the `_local` flag.
        return iter(self.items)

    @dclayMethod(return_=int)
    def __len__(self):
        return len(self.items)

    @dclayMethod(element="anything")
    def append(self, element):
        self.items.append(element)


class StorageList(StorageObject, SplittableCollectionMixin):
    """
    @ClassField chunks list<storageobject>
    """

    @dclayMethod()
    def __init__(self):
        self.chunks = list()

    @dclayMethod(element="anything")
    def append(self, element):
        if self.chunks:
            last_chunk = self.chunks[-1]

            # FIXME: remove this MAGIC NUMBER
            if len(last_chunk) < 500:
                last_chunk.append(element)
                return

        # The collection was either empty or too crowded last element
        # So first prepare a new chunk
        new_chunk = ListChunk([element])

        # add the chunk to the list of chunks
        self.chunks.append(new_chunk)

    @dclayMethod(elements="list<storageobject>")
    def extend(self, elements):
        self.chunks.append(ListChunk(elements))

    @dclayMethod(elements="list<storageobject>", storage_location="anything", _local=True)
    def extend_on_location(self, elements, storage_location):
        chunk = ListChunk(elements)
        chunk.make_persistent(backend_id=storage_location)

        # Required because of the field remoteness
        chunks = self.chunks
        chunks.append(chunk)
        self.chunks = chunks

    @dclayMethod(return_=int)
    def __len__(self):
        return sum(len(chunk) for chunk in self.chunks)

    @dclayMethod(_local=True)
    def __iter__(self):
        from itertools import chain
        return chain(*self.chunks)

    @dclayMethod(return_="anything", idx=int)
    def __getitem__(self, idx):
        """Only integer indexes (no slices) are supported."""
        for chunk in self.chunks:
            l = len(chunk)
            if idx < l:
                return chunk[idx]
            else:
                idx -= l
        raise IndexError("index out of range")


class DictChunk(StorageObject):
    """
    @ClassField elements dict<anything, storageobject>
    """

    @dclayMethod(elements="dict")
    def __init__(self, elements):
        self.elements = {k: v for k, v in elements.items()}

    @dclayMethod(key="anything", return_="anything")
    def __getitem__(self, key):
        return self.elements[key]

    @dclayMethod(key="anything", value="storageobject")
    def __setitem__(self, key, value):
        self.elements[key] = value

    @dclayMethod(key="anything")
    def __delitem__(self, key):
        del self.elements[key]

    @dclayMethod(return_="anything", _local=True)
    def __iter__(self):
        # Note that the following cannot be serialized, hence the `_local` flag.
        return iter(self.elements.keys())

    @dclayMethod(item="anything", return_=bool)
    def __contains__(self, item):
        return item in self.elements

    @dclayMethod(return_="list<anything>")
    def keys(self):
        return list(self.elements.keys())

    @dclayMethod(return_="list<anything>")
    def items(self):
        return self.elements.items()

    @dclayMethod(return_="list<storageobject>")
    def values(self):
        return self.elements.values()

    @dclayMethod(k="anything", return_="storageobject")
    def get(self, k):
        try:
            return self.elements[k]
        except KeyError:
            return None

    @dclayMethod(return_=int)
    def __len__(self):
        return len(self.items)


class StorageDict(StorageObject, SplittableCollectionMixin):
    """
    ToDo: This hash-based dictionary implementation ignores any rebalancing of
    the tree. That doesn't seem right for scalability, but leaving that
    behaviour for the time being.

    @ClassField chunks list<storageobject>
    @ClassField num_buckets int
    """

    @dclayMethod(iterable="anything")
    def __init__(self, iterable):
        # TODO: fix this magic number
        self.num_buckets = 16
        self.chunks = [DictChunk({}) for i in range(self.num_buckets)]
        try:
            data = iterable.items()
        except AttributeError:
            data = iter(iterable)

        for k, v in data:
            self[k] = v

    @dclayMethod(return_="anything", _local=True)
    def __iter__(self):
        # Note that the following cannot be serialized, hence the `_local` flag.
        from itertools import chain
        return chain(iter(c) for c in self.chunks)

    # FIXME: CHECK AND TEST IT (Should change without iter)
    @dclayMethod(return_="anything", _local=True)
    def iteritems(self):
        # Note that the following cannot be serialized, hence the `_local` flag.
        from itertools import chain
        return chain(c.items() for c in self.chunks)

    @dclayMethod(return_="anything", _local=True)
    def iterkeys(self):
        # Note that the following cannot be serialized, hence the `_local` flag.
        from itertools import chain
        return chain(c.keys() for c in self.chunks)

    @dclayMethod(return_="anything", _local=True)
    def itervalues(self):
        # Note that the following cannot be serialized, hence the `_local` flag.
        from itertools import chain
        return chain(c.values() for c in self.chunks)

    @dclayMethod(key="anything", return_="storageobject")
    def __getitem__(self, key):
        h = hash(key) % self.num_buckets
        return self.chunks[h][key]

    @dclayMethod(key="anything", value="storageobject")
    def __setitem__(self, key, value):
        h = hash(key) % self.num_buckets
        self.chunks[h][key] = value

    @dclayMethod(key="anything")
    def __delitem__(self, key):
        h = hash(key) % self.num_buckets
        del self.chunks[h][key]

    @dclayMethod(key="anything", return_="bool")
    def __contains__(self, key):
        h = hash(key) % self.num_buckets
        return key in self.chunks[h]

    @dclayMethod(return_=int)
    def len(self):
        return sum(map(len, self.chunks))

    @dclayMethod(return_="list<anything>")
    def keys(self):
        from itertools import chain
        return list(chain(c.keys() for c in self.chunks))

    @dclayMethod(return_="list<anything>")
    def items(self):
        from itertools import chain
        return list(chain(c.items() for c in self.chunks))

    @dclayMethod(return_="list<storageobject>")
    def values(self):
        from itertools import chain
        return list(chain(c.values() for c in self.chunks))




""" Class description goes here. """

"""Java `properties` parser.

A simple parser of the *.properties files can be found in this module. It is
not a complete and bulletproof implementation, but it is enough for simple
files.

See PropertyFile class for some considerations about the implementation.
"""

from abc import ABCMeta, abstractmethod
import re
import six

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

""" Make this class abstract """


@six.add_metaclass(ABCMeta)
class PropertyFile(object):
    """Abstract property-holder class.

    All Property Files used in dataClay must have it own class, derived from
    this one. This function provides the basic line-by-line iteration and a
    commonruntime interface, but the details on each line stored data is dependant
    on each file.

    WARNING: ** Not implemented **
      - Multiline property lines
      - Escaping sequences
    """

    _prop_comment = re.compile(r"\s*([#!].*)?$")
    _prop_regular_line = re.compile(r"\s*(.*?)\s*[=:]\s*(.*)$")

    def __init__(self, file_name):
        """Open the file (which is expected to be a properties Java file) and read.

        This constructor relies on subclasses implementing their own
        process_line method, which will be called for each line.

        :param file_object: An object-like (stream) for the ".properties" file.
        :return:
        """
        with open(file_name, 'r') as file_object:
            for line in file_object:
                if not self._prop_comment.match(line):
                    m = self._prop_regular_line.match(line)
                    if m is not None:
                        self._process_line(m.group(1), m.group(2))

    @abstractmethod
    def _process_line(self, key, value):
        """Process a line of the ongoing properties file.

        This method should be implemented in derived classes and the internal
        class structure updated according to this properties' file needs.

        :param key: The key for the line being processed.
        :param value: The value (string) for the previous key.
        :return: None
        """
        return


class PropertyDict(PropertyFile):
    """Simple dictionary wrapper for a "properties" file."""

    def __init__(self, file_name):
        super(PropertyDict, self).__init__(file_name)

    def _process_line(self, key, value):
        """Simply store the values in the internal dictionary."""
        self.__dict__[key] = value




""" Class description goes here. """

'''
Created on Jan 25, 2018

@author: dgasull
'''
from weakref import WeakValueDictionary
import logging
from abc import ABCMeta, abstractmethod
import threading
from dataclay.util import Configuration
import six

""" Make this class abstract """


@six.add_metaclass(ABCMeta)
class HeapManager(threading.Thread):
    """
    @summary: This class is intended to manage all dataClay objects in runtime's memory.
    """
    
    """ Logger """
    logger = None
    
    def __init__(self, theruntime):
        """
        @postcondition: Constructor of the object called from sub-class
        @param theruntime: Runtime being managed 
        """ 
        """ Memory objects. This dictionary must contain all objects in runtime memory (client or server), as weakrefs. """
        self.inmemory_objects = WeakValueDictionary()
        threading.Thread.__init__(self)
        self._finished = threading.Event()
        """ Runtime being monitorized. Java uses abstract functions to get the field in the proper type (EE or client) due to type-check. Not needed here. """    
        self.runtime = theruntime
        self.logger = logging.getLogger(__name__)
        self.daemon = True
        self.logger.debug("HEAP MANAGER created.")

    def get_heap(self):
        return self.inmemory_objects
    
    def shutdown(self):
        """Stop this thread"""
        self.logger.debug("HEAP MANAGER shutdown request received.")
        self._finished.set()
        
    def run(self):
        """
        @postcondition: Overrides run function 
        """
        gc_check_time_interval_seconds = Configuration.MEMMGMT_CHECK_TIME_INTERVAL / 1000.0
        while 1:
            self.logger.trace("HEAP MANAGER THREAD is awake...")
            if self._finished.isSet(): break
            self.run_task()
            
            # sleep for interval or until shutdown
            self.logger.trace("HEAP MANAGER THREAD is going to sleep...")
            self._finished.wait(gc_check_time_interval_seconds)
        
        self.logger.debug("HEAP MANAGER THREAD Finished.")
    
    def _add_to_inmemory_map(self, dc_object):
        """
        @postcondition: the object is added to inmemory map
        @param dc_object: object to add
        """
        oid = dc_object.get_object_id()
        self.inmemory_objects[oid] = dc_object
        
    def update_object_id(self, old_object_id, new_object_id):
        """
        @postcondition: Updates heap references when an object changes id
        @param old_object_id: id of the old object id
        @param old_object_id: id of the new object id
        """
        try:
            self.inmemory_objects[new_object_id] = self.inmemory_objects.pop(old_object_id)
        except KeyError:
            self.logger.debug("Miss in Heap object %s" % str(old_object_id))

    def remove_from_heap(self, object_id):
        """
        @postcondition: Remove reference from Heap. Even if we remove it from the heap, 
        the object won't be Garbage collected till HeapManager flushes the object and releases it.
        @param object_id: id of object to remove from heap
        """
        self.inmemory_objects.pop(object_id)

    def get_from_heap(self, object_id):
        """
        @postcondition: Get from heap. 
        @param object_id: id of object to get from heap
        @return Object with id provided in heap or None if not found.
        """
        try:
            obj = self.inmemory_objects[object_id]
            self.logger.debug("Hit in Heap object %s" % str(object_id))
            return obj
        except KeyError:
            self.logger.debug("Miss in Heap object %s" % str(object_id))
            return None

    def exists_in_heap(self, object_id):
        """
        @postcondition: Exists from heap. 
        @param object_id: id of object to get from heap
        @return True if exists. False otherwise.
        """
        try:
            if self.inmemory_objects[object_id] is None:
                return False
            else: 
                return True
        except KeyError:
            return False
  
    def heap_size(self):
        """
        @postcondition: Get heap size. 
        @return Heap size
        """
        return len(self.inmemory_objects)
        
    @abstractmethod
    def flush_all(self): pass
    
    @abstractmethod
    def run_task(self): pass
    
    def cleanReferencesAndLockers(self): 
        """
        @postcondition: Clean references and lockers not being used.
        """ 
        self.runtime.locker_pool.cleanLockers()



"""Common Data Structures for the Storage API.

Note that at the moment (2020-01-09) the classes available in this module are
intended to be used **only** for demonstration purposes.

Please do not use those classes for HPC applications nor any kind of production
applications. However, you can use their API and general behaviour as an
authoritative-ish source for semantics and motivation.
"""




""" Class description goes here. """

from dataclay.util.PropertiesFilesLoader import PropertyFile

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'


class ConfigurationFlags(PropertyFile):
    """Property holder for the "global.properties" file."""

    def __init__(self, property_path):
        super(ConfigurationFlags, self).__init__(property_path)

    def _process_line(self, key, value):
        """The signatures are plain strings."""
        self.__dict__[key] = value




""" Class description goes here. """

"""Global constants holder and discover.

This module is responsible of holding general application-wide constants, and
also provide some basic mechanisms for the Java .properties constant loading.

Note that the package available here (properties) is a placeholder folder which
contains hard links to the properties' files.
"""
from dataclay.exceptions.ErrorDefs import ErrorCodes
from dataclay.util.management.classmgr.PythonTypeSignatures import PythonTypeSignatures

from . import properties

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

PROPERTIES_DIR = properties.__path__[0]

# General global application constants
global_properties = PythonTypeSignatures(PROPERTIES_DIR)

error_codes = ErrorCodes(PROPERTIES_DIR)



"""Common library client API for dataClay.

Runtime.
Note that importing this module has a basic semantic: it prepares the dataClay
core and sets the "client" mode for the library.
"""
import logging.config
import os.path
import sys

from dataclay import getRuntime
from dataclay.DataClayObject import DataClayObject
from dataclay.commonruntime.ClientRuntime import *
from dataclay.commonruntime.ClientRuntime import UNDEFINED_LOCAL as _UNDEFINED_LOCAL
from dataclay.commonruntime.Initializer import initialize, _get_logging_dict_config
from dataclay.communication.grpc.clients.LogicModuleGrpcClient import LMClient
from dataclay.paraver import TRACE_ENABLED, extrae_tracing_is_enabled, get_task_id, \
    set_current_available_task_id
from dataclay.util.StubUtils import track_local_available_classes
from time import sleep

# This will be populated during initialization
LOCAL = _UNDEFINED_LOCAL

__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'
__all__ = ["init", "finish", "DataClayObject"]
logger = logging.getLogger('dataclay.api')
_connection_initialized = False
_initialized = False


def is_initialized():
    """Simple query for the _initialized flag.

    :return: True if `init` has already been called, False otherwise.
    """
    return _initialized


def reinitialize_logging():
    dictconfig = _get_logging_dict_config()

    logger.debug("Ready to close loggers, bye bye!")
    dictconfig["disable_existing_loggers"] = False
    logging.config.dictConfig(dictconfig)
    logger.verbose("Logging reinitialized. Welcome back!")


def reinitialize_clients():
    runtime = getRuntime()
    logger.verbose("Performing reinitialization of clients, removing #%d cached ones and recreating LMClient",
                len(runtime.ready_clients))
    runtime.ready_clients = {
        "@LM": LMClient(settings.logicmodule_host, settings.logicmodule_port),
    }


def init_connection(client_file):
    """Initialize the connection client ==> LogicModule.

    Note that the connection can be initialized standalone from here (like the
    dataClay tool performs) or it can be initialized by the full init() call.

    :param client_file: The path to the `client.properties` file. If set to None,
    then this function assumes that the connection settings are already loaded.
    :return: The LogicModule client (also accessible through the global
    commonruntime.ready_clients["@LM"]
    """
    global _connection_initialized
    logger.debug("Initializing dataClay connection with LM")
    runtime = getRuntime()
    if _connection_initialized:
        logger.warning("Runtime already has a client with the LogicModule, reusing that")
        return runtime.ready_clients["@LM"]

    if client_file:
        settings.load_connection(client_file)

    # Once the properties are load, we can prepare the LM client
    logger.debug("Initializing dataClay connection with LM %s:%s", settings.logicmodule_host, settings.logicmodule_port)
    client = LMClient(settings.logicmodule_host, settings.logicmodule_port)
    runtime.ready_clients["@LM"] = client
    
    _connection_initialized = True

    settings.logicmodule_dc_instance_id = client.get_dataclay_id()

    logger.debug("DataclayInstanceID is %s, storing client in cache", settings.logicmodule_dc_instance_id)
    runtime.ready_clients[settings.logicmodule_dc_instance_id] = runtime.ready_clients["@LM"]

    # wait for 1 python backend 
    while len(get_backends()) < 1:
        logger.info("Waiting for any python backend to be ready ...")
        sleep(2)

    return client


def get_backends():
    """Return all the dataClay backend present in the system."""
    result = getRuntime().get_execution_environments_names("admin", "admin")
    logger.debug("Got %i python backend/s", len(result))
    return result


def get_backends_info():
    """Return all the dataClay BackendInfo present in the system."""
    result = getRuntime().get_execution_environments_info()
    logger.debug("Got %i python backend/s", len(result))
    return result


def get_backend_id_by_name(name):
    """Return dataClay backend present in the system with name provided."""
    all_backends = getRuntime().get_execution_environments_info()
    for backend in all_backends.values():
        if backend.name == name:
            return backend.dataClayID
    return None


def register_dataclay(exthostname, extport):
    """ Register external dataClay for federation
    :param exthostname: external dataClay host name
    :param extport: external dataClay port
    :return: external dataClay ID registered
    :type exthostname: string
    :type extport: int
    :rtype: UUID
    """
    return getRuntime().register_external_dataclay(exthostname, extport)


def get_dataclay_id(exthostname, extport):
    """ Get external dataClay ID with host and port identified
    :param exthostname: external dataClay host name
    :param extport: external dataClay port
    :return: None
    :type exthostname: string
    :type extport: int
    :rtype: None
    """
    return getRuntime().get_external_dataclay_id(exthostname, extport)


def unfederate(ext_dataclay_id=None):
    """ Unfederate all objects belonging to/federated with external data clay with id provided 
    or with all any external dataclay if no argument provided. 
    :param ext_dataclay_id: external dataClay id
    :return: None
    :type ext_dataclay_id: uuid 
    :rtype: None
    """
    if ext_dataclay_id is not None:
        return getRuntime().unfederate_all_objects(ext_dataclay_id)
    else:
        return getRuntime().unfederate_all_objects_with_all_dcs()


def migrate_federated_objects(origin_dataclay_id, dest_dataclay_id):
    """ Migrate federated objects from origin dataclay to destination dataclay
    :param origin_dataclay_id: origin dataclay id 
    :param dest_dataclay_id destination dataclay id
    :return: None
    :rtype: None
    """
    return getRuntime().migrate_federated_objects(origin_dataclay_id, dest_dataclay_id)


def federate_all_objects(dest_dataclay_id):
    """ Federate all objects from current dataclay to destination dataclay
    :param dest_dataclay_id destination dataclay id
    :return: None
    :rtype: None
    """
    return getRuntime().federate_all_objects(dest_dataclay_id)


def pre_network_init(config_file):
    """Perform a partial initialization, with no network."""
    settings.load_properties(config_file)


def init(config_file=None):
    """Initialization made on the client-side, with file-based settings.

    Note that after a successful call to this method, subsequent calls will be
    a no-operation.

    :param config_file: The configuration file that will be used. If not set, then
     the DATACLAYSESSIONCONFIG environment variable will be used and the fallback
     will be the `./cfgfiles/session.properties` path.
    """
    global _initialized

    logger.info("Initializing dataClay API")

    if _initialized:
        logger.warning("Already initialized --ignoring")
        return

    if not config_file:
        # If the call doesn't has config_file set, let's try to fallback into the envvar
        env_config_file = os.getenv("DATACLAYSESSIONCONFIG")
        if env_config_file:
            # If the environment is defined, it is preferred
            config_file = env_config_file
            logger.info("Using the environment variable DATACLAYSESSIONCONFIG=%s",
                        config_file)
        else:
            config_file = "./cfgfiles/session.properties"
            logger.info("Fallback to default ./cfgfiles/session.properties")
    else:
        logger.info("Explicit parameter config_file=\"%s\" will be used",
                    config_file)

    if not os.path.isfile(config_file):
        raise ValueError("dataClay requires a session.properties in order to initialize")

    pre_network_init(config_file)
    post_network_init()


def post_network_init():
    global _initialized

    """Perform the last part of initialization, now with network."""
    client = init_connection(None)

    # In all cases, track (done through babelstubs YAML file)
    contracts = track_local_available_classes()

    # Ensure they are in the path (high "priority")
    sys.path.insert(0, os.path.join(settings.stubs_folder, 'sources'))

    if not contracts:
        logger.warning("No contracts available. Calling new_session, but no classes will be available")

    """ Initialize runtime """
    getRuntime().initialize_runtime()

    session_info = client.new_session(
        settings.current_id,
        settings.current_credential,
        contracts,
        [client.get_dataset_id(settings.current_id, settings.current_credential, dataset) for dataset in settings.datasets],
        client.get_dataset_id(settings.current_id, settings.current_credential, settings.dataset_for_store),
        LANG_PYTHON
    )
    settings.current_session_id = session_info.sessionID

    name = settings.local_backend_name
    if name:
        exec_envs = getRuntime().get_execution_environments_info()
        for k, v in exec_envs.items():
            if exec_envs[k].name == name:
                global LOCAL
                LOCAL = k
                break
        else:
            logger.warning("Backend with name '%s' not found, ignoring", name)

    settings.dataset_id = client.get_dataset_id(
        settings.current_id,
        settings.current_credential,
        settings.dataset_for_store)

    # Remember this function is called after a fork in workers also. 
    # Activate Extrae if needed.
    ### READ #### 
    # Activating tracing with tracing_enabled property set True and starting task id = 0 means we are only tracing dataClay 
    # dataClay client will not increment current available task ID and will send a 0 to LM, which will understand the 0 as 
    # "only dataClay tracing" since for compss it is never 0.
    # Activating tracing with tracing_enabled property set True and starting task id != 0 means we are tracing COMPSs 
    # and dataClay. Current client will not initialize pyextrae or increment task id since COMPSs already initializes 
    # it for us (as a worker). 
    # In any case, none of them needs to add synchronization event or increment the available task id (only services). 
    # Synchronization events are used to merge LM traces and python EE traces. Incrementing available task id is useful to 
    # send to N EE/DS nodes.
    if settings.tracing_enabled:
        logger.info("Initializing tracing")
        # set current available task id 
        set_current_available_task_id(int(settings.extrae_starting_task_id)) 
        # -- Do NOT add synchronization events if it is a compss worker (default case of activation) --
        getRuntime().activate_tracing()
        # TODO: check if is in master, should always be master since initWorker is the one called from workers only?
        # Fallback: if services have tracing active, the call is ignored
        getRuntime().activate_tracing_in_dataclay_services()

    # The new_session RPC may fall, and thus we will consider
    # the library as "not initialized". Arriving here means "all ok".
    _initialized = True


def finish_tracing():
    """
    Finishes tracing if needed 
    """
    if extrae_tracing_is_enabled():
        if int(settings.extrae_starting_task_id) == 0:  # in compss Java runtime will get traces for us
            getRuntime().deactivate_tracing_in_dataclay_services()
            getRuntime().deactivate_tracing()
            getRuntime().get_traces_in_dataclay_services()  # not on workers!
        else:
            getRuntime().deactivate_tracing()

            
def finish():
    global _initialized
    logger.info("Finishing dataClay API")
    finish_tracing()
    getRuntime().stop_runtime()
    _initialized = False

######################################
# Static initialization of dataClay
##########################################################

# The client should never need the delete methods of persistent objects
# ... not doing this is a performance hit
# del DataClayObject.__del__


initialize()


# Now the logger is ready
logger.verbose("Client-mode initialized, dataclay.commonruntime should be ready")




""" Class description goes here. """

from ..LanguageDependantClassInfo import LanguageDependantClassInfo


class PythonClassInfo(LanguageDependantClassInfo):
    _fields = ["id",
               "imports",
               ]




""" Class description goes here. """

from io import BytesIO
import six

from dataclay.serialization.python.DataClayPythonWrapper import DataClayPythonWrapper
from dataclay.serialization.python.lang.BooleanWrapper import BooleanWrapper
from dataclay.serialization.python.lang.IntegerWrapper import IntegerWrapper

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'


class StringWrapper(DataClayPythonWrapper):
    """String with different modes/encodings."""
    __slots__ = ("_mode", "_nullable")

    modes = {"utf-8", "utf-16", "binary"}

    def __init__(self, mode="utf-16", nullable=False):
        assert mode in self.modes, "The String mode should be one in {}".format(self.modes)
        self._mode = mode
        self._nullable = nullable

    def read(self, io_file):
        if self._nullable:
            is_not_null = BooleanWrapper().read(io_file)
            if not is_not_null:
                return None

        size = IntegerWrapper(32).read(io_file)
        ba = io_file.read(size)

        if self._mode == "utf-8":
            return ba.decode('utf-8')
        elif self._mode == "utf-16":
            return ba.decode('utf-16-be')
        elif self._mode == "binary":
            return ba
        else:
            raise TypeError("Internal mode {} not recognized".format(self._mode))

    def write(self, io_file, value):
        if self._nullable:
            if value is None:
                BooleanWrapper().write(io_file, False)
                return
            else:
                BooleanWrapper().write(io_file, True)

        if self._mode == "utf-8":
            ba = value.encode('utf-8')
        elif self._mode == "utf-16":
            ba = value.encode('utf-16-be')
        elif self._mode == "binary":
            if isinstance(value, BytesIO):
                ba = value.getvalue()
            else:
                if six.PY2:
                    ba = bytes(value)
                elif six.PY3:
                    ba = bytes(value, "utf-8")
        else:
            raise TypeError("Internal mode {} not recognized".format(self._mode))

        IntegerWrapper(32).write(io_file, len(ba))
        io_file.write(ba)




""" Class description goes here. """

"""Deactivate dataClay to allow sequential un-modified execution.

This will allow some loaded-modules-mangling and make dataClay classes
harmless.
"""

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2017 Barcelona Supercomputing Center (BSC-CNS)'


class StorageObject(object):
    def make_persistent(self, *args, **kwargs):
        """This becomes a no-op."""
        pass


def dclayMethod(*args, **kwargs):
    return lambda f: f


def deactivate_storage_library():
    """Deactivate the mechanisms of the Storage Library.

    This should be called first thing, and will make sure that StorageObject
    and all that stuff is harmless and a no-op.
    """
    import dataclay
    import storage.api

    # Deactivate those things
    dataclay.DataClayObject = StorageObject
    dataclay.StorageObject = StorageObject
    dataclay.dclayMethod = dclayMethod

    # Also in the storage.api
    storage.api.StorageObject = StorageObject




""" Class description goes here. """

"""High level Storage Interface for dataClay.

This package is used by PyCOMPSs, and uses dataclay.api module.
"""

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'
__version__ = '0.2'





""" Class description goes here. """

'''
Created on 1 feb. 2018

@author: dgasull
'''
from dataclay.loader.DataClayObjectLoader import DataClayObjectLoader
import importlib
from dataclay.commonruntime.ExecutionGateway import ExecutionGateway
from dataclay.serialization.lib.DeserializationLibUtils import DeserializationLibUtilsSingleton


class ClientObjectLoader(DataClayObjectLoader):
    """
    @summary: This class is responsible to create DataClayObjects and load them with data coming from different resources. All possible
    constructions of DataClayObject should be included here. All possible "filling instance" use-cases should be managed here.
    Most lockers should be located here.
    """
    
    def __init__(self, theruntime):
        """
        @postcondition: Constructor of the object 
        @param theruntime: Runtime being managed 
        """ 
        DataClayObjectLoader.__init__(self, theruntime)

    def new_instance(self, class_id, object_id):
        """
        TODO: Refactor this function
        """
        self.logger.verbose("Creating an instance from the class: {%s}", class_id)
    
        try:
            # Note that full_class_name *includes* namespace (Python-specific behaviour)
            full_class_name = self.runtime.local_available_classes[class_id]
        except KeyError:
            raise RuntimeError("Class {%s} is not amongst the locally available classes, "
                               "check contracts and/or initialization" % class_id)
    
        package_name, class_name = full_class_name.rsplit(".", 1)
        m = importlib.import_module(package_name)
        klass = getattr(m, class_name)
        
        return ExecutionGateway.new_dataclay_instance(klass, deserializing=True, object_id=object_id)
    
    def get_or_new_volatile_instance_and_load(self, class_id, object_id, hint, obj_with_data, ifacebitmaps):
        """
        @postcondition: Get from Heap or create a new volatile in EE and load data on it.
        @param class_id: id of the class of the object
        @param object_id: id of the object
        @param hint: hint of the object 
        @param obj_with_data: data of the volatile 
        @param ifacebitmaps: interface bitmaps 
        """
        """
        RACE CONDITION DESIGN
        There are two objects A and B, A -> B, A is persistent and B is volatile.
        There are two threads T1 and T2, T1 is executing a method on A that uses B, when deserializing A, B is loaded into
        heap as a persistent object (all associations are persistent). However, it is actually a volatile send by T2.
        When a volatile server is received and a persistent instance is found, this persistent instance should be "replaced"
        by the new volatile server.
        """
        self.runtime.lock(object_id)
        self.logger.verbose("Get or create new client volatile instance with object id %s in Heap ", str(object_id))
        try:
            """ Double check for race conditions """
            volatile_obj = self.runtime.get_from_heap(object_id)
            if volatile_obj is None: 
                volatile_obj = self.new_instance_internal(class_id, object_id, hint)
                    
            """ Deserialize volatile """ 
            DeserializationLibUtilsSingleton.deserialize_object_with_data_in_client(obj_with_data, volatile_obj, ifacebitmaps, self.runtime,
                                             self.runtime.get_session_id())
            
            # WARNING: RACE CONDITION at EE - during deserialization of volatiles the
            # object may be created and
            # loaded in Heap but not "fully deserialized" yet so even if any execution find
            # it in the
            # heap, object might
            # be not ready (null fields, and no, so is loaded cannot
            # be true till object was fully deserialized)
            volatile_obj.initialize_object_as_volatile()
        finally:
            self.runtime.unlock(object_id)
        
        return volatile_obj




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject
from .Type import Type

    
# Added setterImplementationID, deleted dataclayID
class Property(ManagementObject):
    _fields = ["dataClayID",
               "namespace",
               "className",
               "name",
               "position",
               "type",
               ]

    _internal_fields = ["getterOperationID",
                        "getterImplementationID",
                        "setterImplementationID",
                        "setterOperationID",
                        "updateOperationID",
                        "updateImplementationID",
                        "inMaster",
                        "beforeUpdate",
                        "afterUpdate",
                        "namespaceID",
                        "metaClassID",
                        "languageDepInfos",
                        "annotations"
                        ]

    _typed_fields = {"type": Type}




""" Class description goes here. """

from struct import Struct
from dataclay.serialization.python.DataClayPythonWrapper import DataClayPythonWrapper

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'


class IntegerWrapper(DataClayPythonWrapper):
    """Multiple-size integer type."""
    __slots__ = ("_size", "_type")

    sizes = {8: Struct("!b"),
             16: Struct("!h"),
             32: Struct("!i"),
             64: Struct("!q")}

    def __init__(self, size=32):
        assert size in self.sizes, "Invalid size {:d} for integer type".format(size)
        self._size = size
        self._type = self.sizes[size]

    def read(self, io_file):
        val = io_file.read(int(self._size / 8))
        return self._type.unpack(val)[0]

    def write(self, io_file, value):
        io_file.write(self._type.pack(value))





""" Class description goes here. """

from io import BytesIO
import logging
import lru
import uuid
import traceback
import time
import os
from dataclay.DataClayObject import DataClayObject
from dataclay.commonruntime.Runtime import getRuntime, setRuntime
from dataclay.commonruntime.Runtime import threadLocal
from dataclay.commonruntime.Settings import settings
from dataclay.communication.grpc.clients.ExecutionEnvGrpcClient import EEClient
from dataclay.communication.grpc.clients.LogicModuleGrpcClient import LMClient
from dataclay.communication.grpc.messages.common.common_messages_pb2 import LANG_PYTHON
from dataclay.paraver import set_current_available_task_id, initialize_extrae, finish_tracing, \
    get_traces, extrae_tracing_is_enabled
from dataclay.DataClayObjProperties import DCLAY_GETTER_PREFIX
from dataclay.DataClayObjProperties import DCLAY_PROPERTY_PREFIX
from dataclay.DataClayObjProperties import DCLAY_SETTER_PREFIX
from dataclay.serialization.lib.DeserializationLibUtils import DeserializationLibUtilsSingleton
from dataclay.serialization.lib.SerializationLibUtils import SerializationLibUtilsSingleton
from dataclay.util.FileUtils import deploy_class
from dataclay.util.classloaders import ClassLoader
from dataclay.util.YamlParser import dataclay_yaml_load
from dataclay.commonruntime.ExecutionEnvironmentRuntime import ExecutionEnvironmentRuntime
from ctypes import c_void_p
from dataclay.util import Configuration

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

logger = logging.getLogger(__name__)


class ExecutionEnvironment(object):

    def __init__(self, theee_name):
        self.runtime = ExecutionEnvironmentRuntime(self)
        self.ee_name = theee_name
        """ Initialize runtime """
        self.runtime.initialize_runtime()
            # TODO: de-hardcode this value
        self.cached_sessioninfo = lru.LRU(50)
        self.logger = logging.getLogger(__name__)            
        # This variable will store the following:
        #   - iface_bm (Interface BitMap): used across calls
        #   - session_id (SessionID): is maintained during a deep call
        #   - dataset_id (DataSetID): set on executeImplementation and newPersistentInstance,
        #                             and used for makePersistent of instances.
        self.thread_local_info = threadLocal
        self.init_ee_info()
        
    def init_ee_info(self):
        """
        Initialize EE information (ID). Try to find information in stored files first, otherwise create EE ID. 
        """
        info_file = Configuration.EE_PERSISTENT_INFO_PATH + "infoEE" + self.ee_name
        exists = os.path.isfile(info_file)
        self.logger.info("Reading EE info from %s" % str(info_file))
        if exists:
            fh = open(info_file, 'r+')
            line = fh.readline()
            self.logger.info("READ LINE %s" % str(line))
            self.execution_environment_id = uuid.UUID(line.strip())
            self.logger.info("Initialized EE from file with ID: %s" % str(self.execution_environment_id))
            fh.close()
            
        else:
            self.execution_environment_id = uuid.uuid4()
            self.logger.info("Initialized EE with ID: %s " % str(self.execution_environment_id))

    def store_ee_info(self):
        """
        Store EE information in file 
        """
        info_file = Configuration.EE_PERSISTENT_INFO_PATH + "infoEE" + self.ee_name
        self.logger.info("Storing EE info to %s" % str(info_file))
        exists = os.path.isfile(info_file)
        if not exists:
            fh = open(info_file, 'w')
            self.logger.info("Storing EE info %s" % str(self.execution_environment_id))
            fh.writelines(str(self.execution_environment_id) + "\n")
            fh.close()
            
    def get_execution_environment_id(self):
        """
        Get execution environment id
        :return: execution environment id
        """
        return self.execution_environment_id
        
    def get_runtime(self):
        """
        @return: Runtime of this Execution Environment 
        """ 
        return self.runtime
    
    def prepareThread(self):
        """ 
        Prepare thread local information. Threads contain information about session, dataset,... and it is also used 
        to obtain proper Runtimes. This function was designed for a multithreading design. 
        IMPORTANT: This function should be called at the beginning of all "public" functions in this module.
        """ 
        setRuntime(self.runtime)
    
    def ds_deploy_metaclasses(self, namespace, classes_map_yamls):
        """Deploy MetaClass containers to the Python Execution Environment.
    
        This function stores in a file all the MetaClass, in addition to (optionally)
        putting them into the cache, according to the ConfigOptions.
    
        :param namespace: The namespace 
        :param classes_map: classes map
        :return: The response (empty string)
        """
        try:
            self.prepareThread()
            for class_name, clazz_yaml in classes_map_yamls.items():
                metaclass = dataclay_yaml_load(clazz_yaml)
                ClassLoader.deploy_metaclass_grpc(
                    namespace, class_name, clazz_yaml, metaclass)
        
                if metaclass.name == "UserType" or metaclass.name == "HashType":
                    logger.warning("Ignoring %s dataClay MetaClass", metaclass.name)
                    logger.debug(metaclass)
                    continue
        
                if metaclass.name == "DataClayPersistentObject" \
                        or metaclass.name == "DataClayObject"\
                        or metaclass.name == "StorageObject":
                    continue
        
                logger.info("Deploying class %s to deployment source path %s",
                             metaclass.name, settings.deploy_path_source)

                try:
                    # ToDo: check whether `lang_codes.LANG_PYTHON` or `'LANG_PYTHON'` is the correct key here
                    import_lines = metaclass.languageDepInfos[LANG_PYTHON].imports
                    imports = "\n".join(import_lines)
                except KeyError:
                    # What is most likely is languageDepInfos not having the Python
                    imports = ""
        
                deploy_class(metaclass.namespace, metaclass.name,
                             metaclass.juxtapose_code(True),
                             imports,
                             settings.deploy_path_source,
                             ds_deploy=True)
                logger.info("Deployment of class %s successful", metaclass.name)
        
            return str()
        except:
            traceback.print_exc()
            return str()
        
    def activate_tracing(self, task_id):
        if not extrae_tracing_is_enabled():
            set_current_available_task_id(task_id)
            initialize_extrae(initialize=True)
        
    def deactivate_tracing(self):
        if extrae_tracing_is_enabled():
            finish_tracing(finalize_extrae=True)

    def get_object_metadatainfo(self, object_id):
        """Get the MetaDataInfo for a certain object.
        :param object_id: The ID of the persistent object
        :return: The MetaDataInfo for the given object.
    
        If we have it available in the cache, return it. Otherwise, call the
        LogicModule for it.
        """
        
        self.prepareThread()
        logger.info("Getting MetaData for object {%s}", object_id)
        return getRuntime().get_metadata(object_id)
    
    def get_local_instance(self, object_id, retry=True):
        self.prepareThread()
        return getRuntime().get_or_new_instance_from_db(object_id, retry)
    
    def get_from_db(self, object_id):
        """Get object directly from StorageLocation (DB).
        
        :param session_id: ID of current session
        :param object_id: ID of object to get
        :return: python object
        """
        self.prepareThread()    
        py_object = getRuntime().get_or_new_instance_from_db(object_id, True)
    
        if py_object is None:
            raise Exception("Object from DB returns None")
    
        return py_object
    
    def internal_exec_impl(self, implementation_name, instance, params):
        """Internal (network-agnostic) execute implementation behaviour.
    
        :param instance: The object in which execution will be performed.
        :param implementation_name: Name of the implementation (may also be some dataClay specific $$get)
        :param params: The parameters (args)
        :return: The return value of the function being executed.
        """
        self.prepareThread()
        
        """
        TODO: use better design for this (dgasull) 
        It is possible that a property is set to None by the GC before we 'execute' it. It should be solve by always 
        checking if loaded before returning value. Check race conditions with GC. 
        """
        if not instance.is_loaded(): 
            getRuntime().load_object_from_db(instance, True)
    
        if implementation_name.startswith(DCLAY_GETTER_PREFIX):
            prop_name = implementation_name[len(DCLAY_GETTER_PREFIX):]
            ret_value = getattr(instance, DCLAY_PROPERTY_PREFIX + prop_name)
            logger.debug("Getter: for property %s returned %r", prop_name, ret_value)
            if not isinstance(ret_value, DataClayObject):
                instance.set_dirty(True)
    
        elif implementation_name.startswith(DCLAY_SETTER_PREFIX):
            prop_name = implementation_name[len(DCLAY_SETTER_PREFIX):]
            logger.debug("Setter: for property %s (value: %r)", prop_name, params[0])
            setattr(instance, DCLAY_PROPERTY_PREFIX + prop_name, params[0])
            ret_value = None
            instance.set_dirty(True)
    
        else:
            logger.debug("Call: %s(*args=%s)", implementation_name, params)
            dataclay_decorated_func = getattr(instance, implementation_name)
            ret_value = dataclay_decorated_func._dclay_entrypoint(instance, *params)
    
        return ret_value
    
    def set_local_session(self, session_id):
        """Set the global `self.thread_local_info` with Session.
    
        :param session_id: The UUID for SessionID.
        :return: None
    
        Set the SessionID
        """
        self.prepareThread()
        self.thread_local_info.session_id = session_id
    
    def store_objects(self, session_id, objects_data_to_store, moving, ids_with_alias):
        """
        @postcondition: Store objects in DB 
        @param session_id: ID of session storing objects 
        @param objects_data_to_store: Objects Data to store 
        @param moving: Indicates if store is done during a move 
        @param ids_with_alias: IDs with alias 
        """
        self.prepareThread()
        
        try:
            self.set_local_session(session_id)

        except Exception as e:
            # TODO: Maybe we need to set local session and dataset in some way
            logger.debug("Trying to set_local_session during store of a federated object"
                        "in a federated dataclay ==> Provided dataclayID instead of sessionID")
            pass

        if ids_with_alias is not None:
            for oid in ids_with_alias: 
                self.runtime.add_alias_reference(oid)
    
        for cur_obj_data in objects_data_to_store: 
            object_id = cur_obj_data[0]
            metadata = cur_obj_data[2]
            obj_bytes = cur_obj_data[3]
            # make persistent - session references
            try:
                getRuntime().add_session_reference(object_id)
            except Exception as e:
                # TODO: See exception in set_local_session
                logger.debug("Trying to add_session_reference during store of a federated object"
                            "in a federated dataclay ==> Provided dataclayID instead of sessionID")
                pass

            bytes_for_db = SerializationLibUtilsSingleton.serialize_for_db(object_id, metadata, obj_bytes, True)
            getRuntime().ready_clients["@STORAGE"].store_to_db(settings.environment_id, object_id, bytes_for_db)    
        
    def register_and_store_pending(self, instance, obj_bytes, sync):
        
        object_id = instance.get_object_id()

        # NOTE! We are doing *two* remote calls, and wishlist => they work as a transaction
        getRuntime().ready_clients["@STORAGE"].store_to_db(settings.environment_id, object_id, obj_bytes)    
        # class_id = instance.get_class_extradata().class_id
        # reg_info = [object_id, class_id, instance.get_owner_session_id(), instance.get_dataset_id()]
        
        dataset_id = None
        dcc_extradata = instance.get_class_extradata()
        reg_info = [object_id, dcc_extradata.class_id, instance.get_owner_session_id(), dataset_id]
        try:
            lm_client = getRuntime().ready_clients["@LM"]
            lm_client.register_object(reg_info, settings.environment_id, None, LANG_PYTHON)
        except:
            # do nothing: alias exception
            pass

        instance.set_pending_to_register(False)

        """
        // Inform MDS about new object !
            final Map<ObjectID, MetaClassID> storedObjs = new ConcurrentHashMap<>();
            storedObjs.put(instance.getObjectID(), instance.getMetaClassID());
    
            final Map<ObjectID, SessionID> objsSessions = new ConcurrentHashMap<>();
            objsSessions.put(instance.getObjectID(), instance.getOwnerSessionIDforVolatiles());
            final RegistrationInfo regInfo = new RegistrationInfo(instance.getObjectID(),
                    instance.getMetaClassID(), instance.getOwnerSessionIDforVolatiles(),
                    instance.getDataSetID());
    
            if (DEBUG_ENABLED) {
                logger.debug("[==RegisterPending==] Going to register " + regInfo + " for instance " + System.identityHashCode(instance));
            }
    
            if (sync) {
                final List<RegistrationInfo> regInfos = new ArrayList<>();
                regInfos.add(regInfo);
                this.runtime.getLogicModuleAPI().registerObjects(regInfos,
                        executionEnvironmentID, null, null, Langs.LANG_JAVA);
            } else {
                this.runtime.getLogicModuleAPI().registerObjectsFromDSGarbageCollector(regInfo,
                        executionEnvironmentID,
                        this.runtime);
            }
        """
        
    def store_in_memory(self, session_id, objects_to_store):
        """ This function will deserialize objects into dataClay memory heap using the same design as for
        volatile parameters. Eventually, dataClay GC will collect them, and then they will be
        registered in LogicModule if needed (if objects were created with alias, they must
        have metadata already).
        :param session_id: ID of session of make persistent call
        :param objects_to_store: objects to store.
        :returns: None
        :type session_id: Session ID
        :type objects_to_store: [num_params, imm_objs, lang_objs, vol_params, pers_params]
        :rtype: None
        """
        self.prepareThread()
        self.set_local_session(session_id)

        # No need to provide params specs or param order since objects are not language types
        return DeserializationLibUtilsSingleton.deserialize_params(objects_to_store, None, None, None, getRuntime())
        
    def make_persistent(self, session_id, objects_to_persist):
        """ This function will deserialize make persistent "parameters" (i.e. object to persist
        and subobjects if needed) into dataClay memory heap using the same design as for
        volatile parameters. Eventually, dataClay GC will collect them, and then they will be
        registered in LogicModule if needed (if objects were created with alias, they must
        have metadata already).
        :param session_id: ID of session of make persistent call
        :param objects_to_persist: objects to store.
        :returns: None
        :type session_id: Session ID
        :type objects_to_persist: [num_params, imm_objs, lang_objs, vol_params, pers_params]
        :rtype: None
        """
        logger.debug("Starting make persistent")
        self.store_in_memory(session_id, objects_to_persist)
        logger.debug("Finished make persistent")

    def federate(self, session_id, objects_to_persist):
        """ This function will deserialize object "parameters" (i.e. object to persist
        and subobjects if needed) into dataClay memory heap using the same design as for
        volatile parameters. This function processes objects recieved from federation calls.
        :param session_id: ID of session of federation call
        :param objects_to_persist: objects to store.
        :returns: None
        :type session_id: Session ID
        :type objects_to_persist: [num_params, imm_objs, lang_objs, vol_params, pers_params]
        """

        logger.debug("Starting federation")
        # No need to provide params specs or param order since objects are not language types
        federated_objs = self.store_in_memory(session_id, objects_to_persist)
        for federated_obj in federated_objs:
            try:
                federated_obj.when_federated()
            except:
                # ignore if method is not implemented
                pass
        
        logger.debug("Finished federation")

    def unfederate(self, session_id, object_ids):
        """ This function is called when objects are unfederated.
        :param session_id: ID of session of federation call
        :param object_ids: List of IDs of the objects to unfederate
        :returns: None
        :type session_id: Session ID
        :type objects_to_persist: Object ID
        """
        self.prepareThread()
        self.set_local_session(session_id)
        logger.debug("Starting unfederation: running when_unfederated")
        for object_id in object_ids:
            instance = self.get_local_instance(object_id, True)
            try:
                instance.when_unfederated() 
            except:
                # ignore if method is not implemented
                pass
        
        logger.debug("Finished unfederation")

    def ds_exec_impl(self, object_id, implementation_id, serialized_params_grpc_msg, session_id):
        """Perform a Remote Execute Implementation.
    
        See Java Implementation for details on parameters and purpose.
        """
        self.prepareThread()
        self.set_local_session(session_id)
        logger.debug("Starting new execution")
        logger.debug("SessionID of current execution: %s", session_id)
        logger.debug("ObjectID of current execution: %s", object_id)
        logger.debug("ImplementationID in which the method belongs: %s", implementation_id)
    
        instance = self.get_local_instance(object_id, True)
    
        metaclass_container = instance.get_class_extradata().metaclass_container
        operation = metaclass_container.get_operation(implementation_id)
        logger.debug("DESERIALIZING PARAMETERS")
    
        num_params = serialized_params_grpc_msg[0]
        params = [] 
        if num_params > 0:
            params = DeserializationLibUtilsSingleton.deserialize_params(serialized_params_grpc_msg, None,
                                        operation.params,
                                        operation.paramsOrder, getRuntime())
        logger.debug("STARTING EXECUTION")
    
        ret_value = self.internal_exec_impl(operation.name,
                                       instance,
                                       params)
        
        logger.debug("SERIALIZING RESULT ")
        if ret_value is None:
            logger.debug(" -> Returning None")
            return None
        
        logger.debug(" -> Serializing %s (type: %s)", ret_value, operation.returnType)
        return SerializationLibUtilsSingleton.serialize_params_or_return({0: ret_value},
                                          None,
                                          {"0": operation.returnType},
                                          ["0"],
                                          None,
                                          getRuntime(),
                                          True)  # No volatiles inside EEs
    
    def new_persistent_instance(self, payload):
        """Create, make persistent and return an instance for a certain class."""
    
        raise NotImplementedError("NewPersistentInstance RPC is not yet ready (@ Python ExecutionEnvironment)")
    
    def new_replica(self, session_id, object_id, recursive):
        """Creates a new replica of the object with ID provided in the backend specified.
    
    	:param session_id: ID of session
    	:param object_id: ID of the object
    	:param recursive: Indicates if all sub-objects must be replicated as well.
    	:return: Set of IDs of replicated objects
    	"""
        logger.debug("[==Replica==] New replica for %s", object_id)
        self.prepareThread()
        
        object_ids = set()
        object_ids.add(object_id)
        serialized_objs = self.get_objects(session_id, object_ids, recursive, False)
    
        # Adds associated oid found in metadata in object_ids
        for obj_found in serialized_objs:
            for k in obj_found[2][0]:
                oid = obj_found[2][0][k]
                if oid not in object_ids:
                    logger.verbose("[==Replica==] Associated OID %s found in serialized_objects and not in object_ids", oid)
                    object_ids.add(oid)
                if not recursive:
                    break
    
        logger.debug("[==Replica==] Serialized_objs at the end are %s", serialized_objs)
    
        objects_list = self._check_and_prepare(serialized_objs)
    
        logger.debug("[==Replica==] Object list to store is %s", objects_list)
    
        for obj_id in object_ids:
            for obj in objects_list:
                obj_bytes = obj[3]
                if obj[0] == obj_id:                
                    metadata = obj[2]
                    bytes_for_db = SerializationLibUtilsSingleton.serialize_for_db(obj_id, metadata, obj_bytes, False)
                    getRuntime().ready_clients["@STORAGE"].store_to_db(settings.environment_id, obj_id, bytes_for_db)    
    
        return object_ids
    
    def get_copy_of_object(self, session_id, object_id, recursive):
        """Returns a non-persistent copy of the object with ID provided
    
        :param session_id: ID of session
        :param object_id: ID of the object
        :return: the generated non-persistent objects
        """
        logger.debug("[==Get==] Get copy of %s ", object_id)
        self.prepareThread()
    
        # Get the data service of one of the backends that contains the original object.
        object_ids = set()
        object_ids.add(object_id)
    
        serialized_objs = self.get_objects(session_id, object_ids, recursive, False)
    
        # Prepare OIDs
        logger.debug("[==Get==] Serialized objects obtained to create a copy of %s are %s", object_id, serialized_objs)
        original_to_version = dict()
    
        # Store version in this backend (if already stored, just skip it)
        for obj in serialized_objs:
            orig_obj_id = obj[0]
            version_obj_id = uuid.uuid4()
            original_to_version[orig_obj_id] = version_obj_id
    
        serialized_objs = [self._modify_metadata_oids(obj, original_to_version) for obj in serialized_objs]
        serialized_objs = [self._modify_oid(obj, original_to_version) for obj in serialized_objs]
            
        logger.debug("[==Get==] Serialized Objects after OIDs regeneration are %s", serialized_objs)
            
        i = 0   
        imm_objs = dict()
        lang_objs = dict()
        vol_params = dict()
        pers_params = dict()
        for obj in serialized_objs:
            vol_params[i] = obj
            i = i + 1
    
        serialized_result = [i, imm_objs, lang_objs, vol_params, pers_params]
        
        return serialized_result
        
    def update_object(self, session_id, into_object_id, from_object):
        """Updates an object with ID provided with contents from another object
        :param session_id: ID of session
        :param into_object_id: ID of the object to be updated
        :param from_object: object with contents to be used
        """
        self.prepareThread()
        self.set_local_session(session_id)
        logger.debug("[==PutObject==] Updating object %s", into_object_id)
        object_into = getRuntime().get_or_new_instance_from_db(into_object_id, False)
        object_from = DeserializationLibUtilsSingleton.deserialize_params_or_return(from_object, None, None, None, getRuntime())[0]
        object_into.set_all(object_from)
        logger.debug("[==PutObject==] Updated object %s from object %s", into_object_id, object_from.get_object_id())
        
    def get_objects(self, session_id, object_ids, recursive, moving):
        """Get the serialized objects with id provided
    	:param session_id: ID of session
    	:param object_ids: IDs of the objects to get
    	:param recursive: Indicates if, per each object to get, also obtain its associated objects.
    	:param moving: Indicates we are getting object for a move
    	:return: List of serialized objects with ids provided
        """
        logger.debug("[==Get==] Getting objects %s", object_ids)
        self.prepareThread()
    
        result = list()
        self.thread_local_info.session_id = session_id
        pending_oids_and_hint = list()
        obtained_objs = dict()
        objects_in_other_backend = list()
    
        for oid in object_ids:
            if recursive:
                # Add object to pending
                pending_oids_and_hint.append([oid, None])
                while pending_oids_and_hint:
                    current_oid_and_hint = pending_oids_and_hint.pop()
                    current_oid = current_oid_and_hint[0]
                    current_hint = current_oid_and_hint[1]
                    if current_oid in obtained_objs:
                        # Already Read
                        logger.verbose("[==Get==] Object %s already read", current_oid)
                        continue
    
                    else:
                        try:
                            logger.verbose("[==Get==] Trying to get local instance for object %s", current_oid)    
                            obj_with_data = self.get_object_internal(current_oid, moving)
                            result.append(obj_with_data)
                            obtained_objs[current_oid] = current_hint
                            # Get associated objects and add them to pendings
                            for k in obj_with_data[2][0]:
                                oid_found = obj_with_data[2][0][k]
                                hint_found = obj_with_data[2][2][k]
                                if oid_found != current_oid and oid_found not in obtained_objs:
                                    pending_oids_and_hint.append([oid_found, hint_found])
    
                        except: 
                            logger.debug("[==Get==] Object is in other backend")
                            # Get in other backend
                            objects_in_other_backend.append([current_oid, current_hint])
            else:
                try:
                    result.append(self.get_object_internal(oid, moving))
                except:
                    logger.debug("[==Get==] Object is in other backend")
                    # Get in other backend
                    objects_in_other_backend.append([oid, None])
    
        obj_with_data_in_other_backends = self.get_objects_in_other_backends(session_id, objects_in_other_backend,
                                                                        recursive, moving)
    
        logger.verbose("[==Get==] Object with data return length: %d", len(obj_with_data_in_other_backends))
        logger.trace("[==Get==] Object with data in other backends content: %s", obj_with_data_in_other_backends)
        for obj_in_oth_back in obj_with_data_in_other_backends:
            logger.trace("[==Get==] Append %s to the result", obj_in_oth_back)
            result.append(obj_in_oth_back)
    
        checked_result = self._check_and_prepare(result)
    
        return checked_result
    
    def get_object_internal(self, oid, moving):
        """Get object internal function
    	:param oid: ID of the object ot get
    	:param moving: Indicates we are getting object for a move
    	:return: Object with data
    	:type oid: ObjectID 
    	:type moving: boolean
    	:rtype: Object with data
        """
        # Serialize the object
        logger.verbose("[==GetInternal==] Trying to get local instance for object %s", oid)
        self.prepareThread()
        
        # ToDo: Manage better this try/catch
        getRuntime().lock(oid)  # Race condition with gc: make sure GC does not CLEAN the object while retrieving/serializing it!
        try:
            current_obj = self.get_local_instance(oid, False)
        
            if moving:
                # TODO: Remove hint for moving in order to avoid hints to be also moved.
                pass

            if current_obj.is_pending_to_register():
                # FORCE VOLATILE TO BE REGISTERED
                obj_bytes = SerializationLibUtilsSingleton.serialize_for_db_gc(current_obj, False, None)
                # Register object pending to register now in order to be consistent with any process (metadatas...)
                self.register_and_store_pending(current_obj, obj_bytes, True)
            pending_objs = list()

            # Hint for object located here is here.
            hint = self.get_runtime().get_hint()
    
            # Add object to result and obtained_objs for return and recursive
            return SerializationLibUtilsSingleton.serialize_dcobj_with_data(current_obj, pending_objs,
                                             False, hint, getRuntime(), False)
        finally:
            getRuntime().unlock(oid)          
    
    def get_objects_in_other_backends(self, session_id, objects_in_other_backend, recursive, moving):
        """Get object in another backend. This function is called from DbHandler in a recursive get.
    
    	:param session_id: ID of session
    	:param objects_in_other_backend: List of metadata of objects to read. It is useful to avoid multiple trips.
    	:param recursive: Indicates is recursive
    	:param moving: Indicates if moving
    	:return: ID of objects and for each object, its bytes.
        """
        self.prepareThread()
        result = list()
    
        # Prepare to unify calls (only one call for DS)
        objects_per_backend = dict()
    
        for curr_oid_and_hint in objects_in_other_backend:
            curr_oid = curr_oid_and_hint[0]
            curr_hint = curr_oid_and_hint[1]
            
            logger.debug("[==GetObjectsInOtherBackend==] Looking for metadata of %s", curr_oid)
            
            metadata = self.get_object_metadatainfo(curr_oid)
            logger.info("metadata info are %s", metadata)

            if metadata is None: 
                if curr_hint is not None: 
                    location = curr_hint 
                else: 
                    raise Exception("Object %s not registered" % curr_oid)
            else:
                locations = metadata.locations
                # TODO: Check why always obtain from the first location
                location = next(iter(locations.keys()))
            try:
                objects_in_backend = objects_per_backend[location]
            except KeyError:
                objects_in_backend = set()
                objects_per_backend[location] = objects_in_backend
            objects_in_backend.add(curr_oid)
    
        # Now Call
        for backend_id, objects_to_get in objects_per_backend.items():
    
            logger.debug("[==GetObjectsInOtherBackend==] Get from other location, objects: %s", objects_to_get)
            backend = getRuntime().ready_clients["@LM"].get_executionenvironment_for_ds(backend_id)
    
            try:
                client_backend = getRuntime().ready_clients[backend_id]
            except KeyError:
                logger.verbose("[==GetObjectsInOtherBackend==] Not found Client to ExecutionEnvironment {%s}!"
                               " Starting it at %s:%d", backend_id, backend.hostname, backend.port)
    
                client_backend = EEClient(backend.hostname, backend.port)
                getRuntime().ready_clients[backend_id] = client_backend
    
            cur_result = client_backend.ds_get_objects(session_id, objects_to_get, recursive, moving)
    
            logger.verbose("[==GetObjectsInOtherBackend==] call return length: %d", len(cur_result))
            logger.trace("[==GetObjectsInOtherBackend==] call return content: %s", cur_result)
    
            for res in cur_result:
                result.append(res)
    
        return result
    
    def get_referenced_objects_ids(self, session_id, object_ids):
        """Get ids of objects referenced by objects with IDs provided
        :param session_id: ID of session
        :param object_ids: IDs of the objects that can reference 
        :return: List of object ids referenced by objects with IDs provided
        :type session_id: Session ID
        :type object_ids: set
        :rtype: set of object ids
        """
        logger.debug("[==Get references==] Getting references of objects %s", object_ids)
        self.prepareThread()
        
        result = set()
        self.thread_local_info.session_id = session_id
        pending_oids_and_hint = list()
        obtained_objs = dict()
        objects_in_other_backend = list()
    
        for oid in object_ids:
            # Add object to pending
            pending_oids_and_hint.append([oid, None])
            while pending_oids_and_hint:
                current_oid_and_hint = pending_oids_and_hint.pop()
                current_oid = current_oid_and_hint[0]
                current_hint = current_oid_and_hint[1]
                if current_oid in obtained_objs:
                    # Already Read
                    logger.verbose("[==Get==] Object %s already read", current_oid)
                    continue
    
                else:
                    try:
                        logger.verbose("[==Get==] Trying to get local instance for object %s", current_oid)    
                        obj_with_data = self.get_object_internal(current_oid, False)
                        obtained_objs[current_oid] = current_hint
                        # Get associated objects and add them to pendings
                        for k in obj_with_data[2][0]:
                            oid_found = obj_with_data[2][0][k]
                            hint_found = obj_with_data[2][2][k]
                            if oid_found != current_oid and oid_found not in obtained_objs:
                                pending_oids_and_hint.append([oid_found, hint_found])
                                result.add(oid_found)
    
                    except Exception: 
                        logger.debug("[==Get==] Object is in other backend")
                        # Get in other backend
                        objects_in_other_backend.append([current_oid, current_hint])
    
        oids_in_other_ee = self.get_objects_ids_in_other_backends(session_id, objects_in_other_backend)
        for other_ee_oid in oids_in_other_ee:
            result.add(other_ee_oid)
            
        logger.debug("[==Get references==] Result %s", str(result))
   
        return result
    
    def get_objects_ids_in_other_backends(self, session_id, objects_in_other_backend):
        """Get IDs of objects referenced by provided object ids in another backend
        :param session_id: ID of session
        :param objects_in_other_backend: List of metadata of objects to read. It is useful to avoid multiple trips.
        :return: ID of objectss in another backend
        :type session_id: Session ID
        :type object_ids: set
        :rtype: set of object ids referenced by provided object ids in another backend
        """ 
        self.prepareThread()
        result = list()
    
        # Prepare to unify calls (only one call for DS)
        objects_per_backend = dict()
    
        for curr_oid_and_hint in objects_in_other_backend:
            curr_oid = curr_oid_and_hint[0]
            curr_hint = curr_oid_and_hint[1]
            logger.debug("[==GetOIDsObjectsInOtherBackend==] Looking for metadata of %s", curr_oid)
            metadata = self.get_object_metadatainfo(curr_oid)
            logger.info("metadata info are %s", metadata)
            if metadata is None: 
                if curr_hint is not None: 
                    location = curr_hint 
                else: 
                    raise Exception("Object %s not registered" % curr_oid)
            else:
                locations = metadata.locations
                # TODO: Check why always obtain from the first location
                location = next(iter(locations.keys()))
    
            try:
                objects_in_backend = objects_per_backend[location]
            except KeyError:
                objects_in_backend = set()
                objects_per_backend[location] = objects_in_backend
            objects_in_backend.add(curr_oid)
    
        # Now Call
        for backend_id, objects_to_get in objects_per_backend.items():
    
            logger.debug("[==GetOIDsObjectsInOtherBackend==] Get from other location, objects: %s", objects_to_get)
            backend = getRuntime().ready_clients["@LM"].get_executionenvironment_for_ds(backend_id)
    
            try:
                client_backend = getRuntime().ready_clients[backend_id]
            except KeyError:
                logger.verbose("[==GetOIDsObjectsInOtherBackend==] Not found Client to ExecutionEnvironment {%s}!"
                               " Starting it at %s:%d", backend_id, backend.hostname, backend.port)
    
                client_backend = EEClient(backend.hostname, backend.port)
                getRuntime().ready_clients[backend_id] = client_backend
    
            cur_result = client_backend.get_referenced_objects_ids(session_id, objects_to_get)
    
            logger.verbose("[==GetOIDsObjectsInOtherBackend==] call return length: %d", len(cur_result))
            logger.trace("[==GetOIDsObjectsInOtherBackend==] call return content: %s", cur_result)
    
            for res in cur_result:
                result.append(res)
    
        return result
    
    def new_version(self, session_id, object_id, metadata_info):
        """Creates a new version of the object with ID provided in the backend specified.
    
    	:param session_id: ID of session
    	:param object_id: ID of the object
    	:param metadata_info: Metadata of the object, including the backends where the root object to be versioned is located
    	:return: The OID of the version root and the mapping from version OID to original OID for each versioned object
        """
        self.prepareThread()
        logger.debug("[==Version==] New version for %s ", object_id)
    
        # Get the data service of one of the backends that contains the original object.
        object_ids = set()
        object_ids.add(object_id)
    
        serialized_objs = self.get_objects(session_id, object_ids, True, False)
    
        # Prepare OIDs
        logger.debug("[==Version==] Serialized objects obtained to create version for %s are %s", object_id, serialized_objs)
        version_to_original = dict()
        original_to_version = dict()
        versions_hints = dict()
    
        # Store version in this backend (if already stored, just skip it)
        for obj in serialized_objs:
            orig_obj_id = obj[0]
            version_obj_id = uuid.uuid4()
            version_to_original[version_obj_id] = orig_obj_id
            original_to_version[orig_obj_id] = version_obj_id
            # ToDo: Manage hints for versions_hints dict
            # versions_hints[version_obj_id] = hint
    
        # ToDo: Add also versions_hints to modify_metadata_oids
        serialized_objs = [self._modify_metadata_oids(obj, original_to_version) for obj in serialized_objs]
        serialized_objs = [self._modify_oid(obj, original_to_version) for obj in serialized_objs]
    
        logger.debug("[==Version==] Serialized Objects after modification are %s", serialized_objs)
    
        # Store versions
        objs_to_store = self._check_and_prepare(serialized_objs)
    
        for obj in objs_to_store:
            obj_id = obj[0]
            metadata = obj[2]
            obj_bytes = obj[3]
            logger.info("obj_id is %s, obj_data are %s", obj_id, obj_bytes)
            
            bytes_for_db = SerializationLibUtilsSingleton.serialize_for_db(obj_id, metadata, obj_bytes, False)
            
            getRuntime().ready_clients["@STORAGE"].store_to_db(settings.environment_id, obj_id, bytes_for_db)
            
        # Modify metadata_info
        version_id = original_to_version[object_id]
        environments = dict()
        environments[settings.environment_id] = getRuntime().ready_clients["@LM"].get_executionenvironment_for_ds(settings.environment_id)
    
        logger.debug("[==Version==] Modifying metadataInfo: %s", metadata_info)
        version_md = metadata_info
        version_md.locations = environments
        logger.debug("[==Version==] Added metadata info to MetaData Cache: %s : %s", version_id, version_md)
        getRuntime().metadata_cache[version_id] = version_md
        logger.debug("[==Version==] Version finished for object %s , newVersion oid is: %s", object_id, version_id)
    
        return version_id, version_to_original
    
    def consolidate_version(self, session_id, version):
        """Consolidates all the objects in versionInfo, being the current data service the one containing all the versioned
    	   objects. For each versioned object, its OID is set to the original one according to the mapping in versionInfo,
    	   and the consolidated object is stored in the same locations as the original one (before versioning). The versions
    	   are deleted.
    
    	:param session_id:ID of session
    	:param version: Info of the version
        """
        self.prepareThread()
        # Consolidate in this backend - the complete version is here
        version_to_original = version.versionsMapping
        original_hints = dict()

        for version_oid, original_oid in version_to_original.items():
            original_md = version.originalMD[original_oid]
            original_hints[original_oid] = list(original_md.locations.keys())[0]

        logger.debug("[==Consolidate==] Consolidating version %s to original %s", version.versionOID, version_to_original.get(version.versionOID))
    
        version_object_ids = set()

        # Get bytes of all version objects
        for k in list(version_to_original.keys()):
            version_object_ids.add(k)

        dirty_version_bytes = self.get_objects(session_id, version_object_ids, True, False)
        logger.debug("[==Consolidate==] Version objs obtained are %s", dirty_version_bytes)
    
        # Update original objects
        # ToDo: Change also versions_hints in modify_metadata_oids
        version_bytes = list()
    
        for vers_byte in dirty_version_bytes:
            # Change the stream position to 0
            vers_byte[3].seek(0)
            # Modify metadata and oid with the versions one
            modified_metadata = self._modify_metadata_oids(vers_byte, version_to_original)
            modified_oid = self._modify_oid(modified_metadata, version_to_original)
            version_bytes.append(modified_oid)
    
        logger.trace("[==Consolidate==] Version objs modified are %s", version_bytes)
    
        dest_loc = None
    
        try:
            self.thread_local_info.session_id = session_id
    
            # Update original objects (here and in other DSs - replicas)
            orig_oid = version_to_original[version.versionOID]
            md_info = version.originalMD[orig_oid]
            locs = md_info.locations
            for loc_id in locs:
                dest_loc = locs[loc_id]
                if loc_id == settings.environment_id:
                    logger.verbose("[==Consolidate==] Upsert Objects in this DS")
                    self.upsert_objects(session_id, version_bytes)
                else:
                    try:
                        st_client = getRuntime().ready_clients[loc_id]
                    except KeyError:
                        logger.verbose("[==Consolidate==] Not found Client to ExecutionEnvironment {%s}!"
                                       " Starting it at %s:%d", loc_id, dest_loc.hostname, dest_loc.port)
    
                        st_client = EEClient(dest_loc.hostname, dest_loc.port)
                        getRuntime().ready_clients[loc_id] = st_client
    
                    logger.debug("[==Consolidate==] Going to other DS to upsert %s", version_to_original[version.versionOID])
                    st_client.ds_upsert_objects(session_id, version_bytes)
    
                # Delete versions here
                # TODO: G.C. should do it
                # commonruntime.ready_clients["@STORAGE"].ds_remove_objects(session_id, version_object_ids, True, False, dest_loc)
    
        except Exception as e:
            traceback.print_exc()
            logger.error("Exception during Consolidate Version")
            raise e
        
        logger.debug("[==Consolidate==] Consolidate ended")
    
    def _check_and_prepare(self, serialized_objs):
        """ Check if serialized_obj are in the right format and prepare the objects_list to send to store_objects """
        objects_list = list()
    
        for obj in serialized_objs:
            logger.trace("[==CheckAndPrepare==] Obj in Serialized objs content: %s", obj)
    
            # Since the ds_store_objects is serializing again we deserialize the data before sending it
            # ToDo: Correct this
            if not isinstance(obj[3], BytesIO):
                logger.verbose("Type for (obj[3]): %s", type(obj))
                logger.trace("Bytes (obj[3]) serialized in wrong way for %s", obj)
                serialized_obj = BytesIO(obj[3])
            else:
                serialized_obj = obj[3]
            dcobj = obj[0], obj[1], obj[2], serialized_obj
            objects_list.append(dcobj)
    
        return objects_list
    
    def _modify_metadata_oids(self, obj, original_to_version):
        """ Modify the version's metadata in serialized_objs with original OID"""
        self.prepareThread()
        logger.debug("[==ModifyMetadataOids==] Modify serialized object %r", obj)
        logger.debug("[==ModifyMetadataOids==] Version OIDs Map: %s", original_to_version)

        for tag, oid in obj[2][0].items():
            try:
                obj[2][0][tag] = original_to_version[oid]
            except KeyError:
                logger.debug("[==ModifyMetadataOids==] oid %s is not mapped => object added in the version", oid)
                # obj[2][0][tag] = oid
                pass

        logger.debug("[==ModifyMetadataOids==] Object with modified metadata is %r", obj)

        return obj
    
    def _modify_oid(self, obj, original_to_version):
        """ Modify the version's OID in serialized_objs with the original OID"""
        try:
            new_obj = original_to_version[obj[0]], obj[1], obj[2], obj[3]
        except KeyError:
            # OID in object are already changed
            new_obj = obj
        return new_obj

    def upsert_objects(self, session_id, object_ids_and_bytes):
        """Updates objects or insert if they do not exist with the values in objectBytes.
        NOTE: This function is recursive, it is going to other DSs if needed.
        :param session_id: ID of session needed.
    	:param object_ids_and_bytes: Map of objects to update.
    	"""
        self.prepareThread()
        self.thread_local_info.session_id = session_id

        try:
            objects_in_other_backends = list()
    
            # To check for replicas
            for cur_entry in object_ids_and_bytes:
                # ToDo: G.C. stuffs
                object_id = cur_entry[0]
                logger.debug("[==Upsert==] Updated or inserted object %s", object_id)
                try:
                    # Update bytes at memory object
                    logger.debug("[==Upsert==] Getting/Creating instance from upsert with id %s", object_id)
                    instance = getRuntime().get_or_new_instance_from_db(object_id, False)
                    DeserializationLibUtilsSingleton.deserialize_object_with_data(cur_entry, instance, None, getRuntime(), getRuntime().get_session_id(), True)
                    
                    instance.set_dirty(True)
                    
                except Exception:
                    # Get in other backend
                    objects_in_other_backends.append(cur_entry)
    
            self.upsert_objects_in_other_backend(session_id, objects_in_other_backends)
    
        except Exception as e:
            traceback.print_exc()
            logger.error("Exception during Upsert Objects")
            raise e
    
    def upsert_objects_in_other_backend(self, session_id, objects_in_other_backends):
        """Update object in another backend.
    
    	:param session_id: ID of session
    	:param objects_in_other_backends: List of metadata of objects to update and its bytes. It is useful to avoid multiple trips.
    	:return: ID of objects and for each object, its bytes.
        """
        self.prepareThread()
        # Prepare to unify calls (only one call for DS)
        objects_per_backend = dict()
        for curr_obj_with_ids in objects_in_other_backends:
            
            curr_oid = curr_obj_with_ids[0]
            locations = self.get_object_metadatainfo(curr_oid).locations
            location = next(iter(locations.keys()))
            # Update object at first location (NOT UPDATING REPLICAS!!!)
            try:
                objects_in_backend = objects_per_backend[location]
            except KeyError:
                objects_in_backend = list()
                objects_per_backend[location] = objects_in_backend
                
            objects_in_backend.append(curr_obj_with_ids)
        # Now Call
        for backend_id, objects_to_update in objects_per_backend.items():
    
            backend = getRuntime().ready_clients["@LM"].get_executionenvironment_for_ds(backend_id)
    
            try:
                client_backend = getRuntime().ready_clients[backend_id]
            except KeyError:
                logger.verbose("[==GetObjectsInOtherBackend==] Not found Client to ExecutionEnvironment {%s}!"
                               " Starting it at %s:%d", backend_id, backend.hostname, backend.port)
    
                client_backend = EEClient(backend.hostname, backend.port)
                getRuntime().ready_clients[backend_id] = client_backend
    
            client_backend.ds_upsert_objects(session_id, objects_to_update)

    def move_objects(self, session_id, object_id, dest_backend_id, recursive):
        """This operation removes the objects with IDs provided NOTE:
         This function is recursive, it is going to other DSs if needed.
    
    	:param session_id: ID of session.
    	:param object_id: ID of the object to move.
        :param dest_backend_id: ID of the backend where to move.
        :param recursive: Indicates if all sub-objects (in this location or others) must be moved as well.
    	:return: Set of moved objects.
        """
        update_metadata_of = set()

        try:
            logger.debug("[==MoveObjects==] Moving object %s to storage location: %s", object_id, dest_backend_id)
            object_ids = set()
            object_ids.add(object_id)

            # TODO: Object being used by session (any oid in the method header) G.C.

            serialized_objs = self.get_objects(session_id, object_ids, recursive, True)
            objects_to_remove = set()
            objects_to_move = list()

            for obj_found in serialized_objs:
                logger.debug("[==MoveObjects==] Looking for metadata of %s", obj_found[0])
                metadata = self.get_object_metadatainfo(obj_found[0])
                obj_location = list(metadata.locations.keys())[0]

                if obj_location == dest_backend_id:
                    logger.debug("[==MoveObjects==] Ignoring move of object %s since it is already where it should be."
                    " ObjLoc = %s and DestLoc = %s" , obj_found[0], obj_location, dest_backend_id)
					
					# object already in dest
                    pass
                else:
                    if settings.storage_id == dest_backend_id:
                        # THE DESTINATION IS HERE
                        if obj_location != settings.storage_id:
                            logger.debug("[==MoveObjects==] Moving object  %s since dest.location is different to src.location and object is not in dest.location."
										 " ObjLoc = %s and DestLoc = %s", obj_found[0], obj_location, dest_backend_id)
                            objects_to_move.append(obj_found)
                            objects_to_remove.add(obj_found[0])
                            update_metadata_of.add(obj_found[0])
                        else:
                            logger.debug("[==MoveObjects==] Ignoring move of object %s since it is already where it should be"
										 " ObjLoc = %s and DestLoc = %s", obj_found[0], obj_location, dest_backend_id)
                    else:
                        logger.debug("[==MoveObjects==] Moving object %s since dest.location is different to src.location and object is not in dest.location "
                                        " ObjLoc = %s and DestLoc = %s", obj_found[0], obj_location, dest_backend_id)
                        # THE DESTINATION IS ANOTHER NODE: move.
                        objects_to_move.append(obj_found)
                        objects_to_remove.add(obj_found[0])
                        update_metadata_of.add(obj_found[0])

            logger.debug("[==MoveObjects==] Finally moving OBJECTS: %s", objects_to_remove)

            try:
                sl_client = getRuntime().ready_clients[dest_backend_id]
            except KeyError:
                st_loc = getRuntime().get_execution_environments_info()[dest_backend_id] 
                self.logger.debug("Not found in cache ExecutionEnvironment {%s}! Starting it at %s:%d",
                               dest_backend_id, st_loc.hostname, st_loc.port)
                sl_client = EEClient(st_loc.hostname, st_loc.port)
                getRuntime().ready_clients[dest_backend_id] = sl_client

            sl_client.ds_store_objects(session_id, objects_to_move, True, None)

            # TODO: lock any execution in remove before storing objects in remote dataservice so anyone can modify it.
			# Remove after store in order to avoid wrong executions during the movement :)
			# Remove all objects in all source locations different to dest. location
            # TODO: Check that remove is not necessary (G.C. Should do it?)
            # getRuntime().ready_clients["@STORAGE"].ds_remove_objects(session_id, object_ids, recursive, True, dest_backend_id)

            for oid in objects_to_remove:
                if oid in getRuntime().metadata_cache:
                    del getRuntime().metadata_cache[oid]
            logger.debug("[==MoveObjects==] Move finalized ")

        except Exception as e:
            logger.error("[==MoveObjects==] Exception %s", e.args)

        return update_metadata_of
    
    def update_refs(self, ref_counting):
        """ forward to SL """ 
        self.prepareThread()
        getRuntime().ready_clients["@STORAGE"].update_refs(ref_counting)    

    def get_retained_references(self):
        return self.runtime.get_retained_references()
    
    def close_session_in_ee(self, session_id):
        self.runtime.close_session_in_ee(session_id)
        
    def exists(self, object_id):
        
        # object might be in heap but as a "proxy" 
        # since this function is used from SL after checking if the object is in database, 
        # we return false if the object is not loaded so the combination of SL exists and EE exists
        # can tell if the object actually exists
        # summary: the object only exist in EE if it is loaded. 
        
        in_heap = self.runtime.exists(object_id)
        if in_heap: 
            obj = self.runtime.get_from_heap(object_id)
            return obj.is_loaded()
        else:
            return False

    def get_federated_objects(self, ext_dataclay_id, object_ids):
        
        self.prepareThread()
        logger.debug("[==GetFederatedObjects==] Getting objects %s", object_ids)
        result = []
        try:
            for object_id in object_ids:
                if not getRuntime().ready_clients["@LM"].check_object_is_federated_with_dataclay_instance(
                    object_id, ext_dataclay_id):
                    
                    logger.debug("[==GetFederatedObjects==] Object %s not federated with ext.dataclay %s", object_id, ext_dataclay_id)

                    return None
            
            for object_id in object_ids:
                obj_with_data = self.get_object_internal(object_id, False)
                result.append(obj_with_data)
        
        except Exception as e:
            traceback.print_exc()
            logger.error("[==GetFederatedObjects==] Exception %s", e)
        
        return result

    def get_traces(self):
        logger.debug("Merging...")
        return get_traces()
    
    ################## EXTRAE IGNORED FUNCTIONS ###########################
    deactivate_tracing.do_not_trace = True
    activate_tracing.do_not_trace = True




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class SessionDataContract(ManagementObject):
    _fields = ["id",
               "dataContractID",
               "dataSetOfProvider",
               ]

    _internal_fields = []




""" Class description goes here. """

from itertools import starmap

from jinja2 import Template

from dataclay.serialization.python.lang.BooleanWrapper import BooleanWrapper
from dataclay.serialization.python.lang.FloatWrapper import FloatWrapper
from dataclay.serialization.python.lang.IntegerWrapper import IntegerWrapper
from dataclay.serialization.python.lang.NullWrapper import NullWrapper
from .Type import Type

NATIVE_PACKAGES = {
    'numpy',
    'caffe',
    'csv',
}

STATIC_ATTRIBUTE_FOR_EXTERNAL_INIT = 'DCLAY_FORCE_EXTERNAL_INIT'

# Static Template for the source code of the classes
py_code = Template("""
class {{ class_name }}({{ parent_name }}):
    \"\"\"Auto-generated code for class {{ metaclass.name }}

    This source code has been generated by the dataclay MetaClass container. There is
    some work ToDo yet.
    \"\"\"
{% for c in imp_codes %}
{{ c }}{% endfor %}
""")

stub_only_def = Template("""
    @dclayEmptyMethod
    def {{ func_name }}(
            self{% for param in param_names %}{% if loop.first %},{% endif %}
            {{ param }}{% if loop.last %}
    {% endif %}{% else %}
    {% endfor %}):
        raise NotImplementedError("Language Error: Method {{ func_name }} is not available for Python")
""")

# Note that the class_id of language types are null since "dataClay 2"
mapping_table = [
    (("int", int, IntegerWrapper(64)), Type(
        signature='J',
        includes=[],
    )),
    (("float", float, FloatWrapper(64)), Type(
        signature='D',
        includes=[],
    )),
    (("bool", bool, BooleanWrapper()), Type(
        signature='Z',
        includes=[],
    )),
    (("None", None, NullWrapper()), Type(
        signature='V',
        includes=[],
    )),
]

# Statically build the dictionaries for fast and easy lookup
docstring_types = dict(starmap(
    lambda what, type_c: (what[0], type_c),
    mapping_table))

instance_types = dict(starmap(
    lambda what, type_c: (what[1], type_c),
    mapping_table))

serialization_types = dict(starmap(
    lambda what, type_c: (type_c.signature, what[2]),
    mapping_table
))




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject
from .Type import Type


class Operation(ManagementObject):
    _fields = ["dataClayID",
               "namespace",
               "className",
               "descriptor",
               "signature",
               "name",
               "nameAndDescriptor",
               "params",
               "paramsOrder",
               "returnType",
               "implementations",
               "isAbstract",
               "isStaticConstructor",
               ]

    _internal_fields = ["metaClassID",
                        "namespaceID",
                        "languageDepInfos",
                        "annotations"
                        ]

    _typed_fields = {"returnType": Type}




""" Class description goes here. """

import os.path

from dataclay.util.PropertiesFilesLoader import PropertyFile

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'


class PythonTypeSignatures(PropertyFile):
    """Property holder for the "python_type_signatures.properties" file."""

    def __init__(self, property_path, python_type_name='python_type_signatures.properties'):
        super(PythonTypeSignatures, self).__init__(os.path.join(
            property_path, python_type_name))

    def _process_line(self, key, value):
        """The signatures are plain strings."""
        self.__dict__[key] = value




""" Class description goes here. """

'''
Created on 2 ene. 2018

@author: daniel
'''

'''

WARNING: THIS CLASS IS MANTAINED ONLY BECAUSE MF2C DEMO IS IMPORTING MODULES FROM HERE AND IT IS WRONG!

'''

from dataclay.DataClayObjMethods import dclayMethod
from dataclay.DataClayObject import DataClayObject




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class InterfaceInContract(ManagementObject):
    _fields = ["id",
               "iface",
               "implementationsSpecPerOperation",
               ]

    _internal_fields = ["interfaceID",
                        "accessibleImplementations",
                        ]




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject

    
class ExecutionEnvironment(ManagementObject):
    _fields = ["dataClayID",
               "hostname",
               "name",
               "port",
               "lang"
               ]

    _internal_fields = []




""" Class description goes here. """

"""Interceptor for GRPC server calls."""

from grpc import ServerInterceptor
from logging import getLogger

from dataclay import PrvManager
from dataclay.paraver.prv_traces import TraceType
from dataclay.commonruntime.Settings import settings
from . import HEADER_MESSAGEID
from decorator import decorator, decorate
import time

logger = getLogger(__name__)


class ParaverServerInterceptor(ServerInterceptor):

    request_id = None
    sv_client_port = None

    def __init__(self, origin_hostname, origin_port):
        logger.debug("Initialize ParaverServerInterceptor")
        self.origin_hostname = origin_hostname
        self.origin_port = origin_port
        self.prv_manager = PrvManager.get_manager()

    def intercept_service(self, continuation, handler_call_details):
        """
        Override of the ServerInterceptor method.
        If paraver is active (activate_paraver_traces) get the req_id from metadata
        and store the client_port as class var, then add the receive trace to the traces queue.

        :param continuation: A function that takes a HandlerCallDetails and proceeds to invoke the next interceptor in the chain,
                             if any, or the RPC handler lookup logic, with the call details passed as an argument,
                             and returns an RpcMethodHandler instance if the RPC is considered serviced, or None otherwise.
        :param handler_call_details: A HandlerCallDetails describing the RPC.
        :return: Response of the RPC.
        """
        if settings.paraver_tracing_enabled:

            logger.debug("Intercept")
            
            try:
                ParaverServerInterceptor.request_id = int(handler_call_details.invocation_metadata[1].value)
            except Exception:
                try:
                    ParaverServerInterceptor.request_id = int(handler_call_details.invocation_metadata[0].value)
                except Exception:
                    return continuation(handler_call_details)

            if "java" in handler_call_details.invocation_metadata[0].value:
                # If client call comes from Java client_port is logicmodule_port
                ParaverServerInterceptor.sv_client_port = settings.logicmodule_port
                
            elif "python" in handler_call_details.invocation_metadata[1].value:
                # If comes from python client_port is static one defined in client interceptor
                ParaverServerInterceptor.sv_client_port = 892892
                
            client_port = ParaverServerInterceptor.sv_client_port

            self.prv_manager.add_network_receive(
                self.origin_hostname,
                client_port, 
                ParaverServerInterceptor.request_id,
                0 # unknown/unused method id
            )
            response = continuation(handler_call_details)

            return response

        else:
            return continuation(handler_call_details)

    def add_send(self):
        """ If Paraver is active add the send trace to the traces queue"""
        if settings.paraver_tracing_enabled:
            if ParaverServerInterceptor.request_id == None:
                # Interceptor didn't receive request
                return
            
            self.prv_manager.add_network_send(
                int(time.time() * 1000000000),
                TraceType.SEND_RESPONSE,
                self.origin_port,
                ParaverServerInterceptor.request_id,
                self.origin_hostname,
                ParaverServerInterceptor.sv_client_port,
                0,  # unknown/unused message size
                0   # unknown/unused method id
            )




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class SessionContract(ManagementObject):
    _fields = ["id",
               "contractID",
               "sessionInterfaces",
               ]

    _internal_fields = []




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject

    
class DataClayInstance(ManagementObject):
    _fields = ["dcID",
               "hosts",
               "ports",
               ]

    _internal_fields = []




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class OpImplementations(ManagementObject):
    _fields = ["id",
               "operationSignature",
               "numLocalImpl",
               "numRemoteImpl",
               ]

    _internal_fields = ["localImplementationID",
                        "remoteImplementationID",
                        ]



"""Management of Python Classes.

This module is responsible of management of the Class Objects. A central Python
Metaclass is responsible of Class (not object) instantiation.

Note that this managers also includes most serialization/deserialization code
related to classes and function call parameters.
"""
from decorator import getfullargspec
import inspect
from logging import getLogger
import six

from dataclay.util.management.classmgr.MetaClass import MetaClass
from dataclay.util.management.classmgr.Operation import Operation
from dataclay.util.management.classmgr.Property import Property
from dataclay.util.management.classmgr.Type import Type
from dataclay.util.management.classmgr.Utils import STATIC_ATTRIBUTE_FOR_EXTERNAL_INIT
from dataclay.util.management.classmgr.python.PythonClassInfo import PythonClassInfo
from dataclay.util.management.classmgr.python.PythonImplementation import PythonImplementation
from dataclay.exceptions.exceptions import DataClayException

# Publicly show the dataClay method decorators
__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2016 Barcelona Supercomputing Center (BSC-CNS)'

logger = getLogger(__name__)

# Populated by ExecutionGateway, can be freely accessed and cleared (see MetaClassFactory)
loaded_classes = set()

# Class extradata caches
class_extradata_cache_exec_env = dict()
class_extradata_cache_client = dict()


class ExecutionGateway(type):
    """Python' Metaclass dataClay "magic"

    This type-derived Metaclass is used by DataClayObject to control the class
    instantiation and also object instances.
    """

    def __new__(mcs, classname, bases, dct):

        if classname == 'DataClayObject':
            # Trivial implementation, do nothing
            return super(ExecutionGateway, mcs).__new__(mcs, classname, bases, dct)
        # at this point, a real dataClay class is on-the-go
        klass = super(ExecutionGateway, mcs).__new__(mcs, classname, bases, dct)
        loaded_classes.add(klass)
        return klass

    def __init__(cls, name, bases, dct):
        logger.verbose("Initialization of class %s in module %s",
                       name, cls.__module__)

        super(ExecutionGateway, cls).__init__(name, bases, dct)

    def __call__(cls, *args, **kwargs):
        if cls.__name__ == 'DataClayObject':
            raise TypeError("Cannot create base objects")
        
        if getattr(cls, STATIC_ATTRIBUTE_FOR_EXTERNAL_INIT, False):
            logger.debug("New Persistent Instance (remote init) of class `%s`", cls.__name__)
            raise NotImplementedError("External initialization not implemented")
            # return getRuntime().new_persistent_instance_aux(cls, args, kwargs)
        else:
            ret = object.__new__(cls)  # this defers the __call__ method
            logger.debug("New regular dataClay instance of class `%s`", cls.__name__)

            # The following will trigger the initialize_object,
            # which will take care of volatile (in Execution Environment, ofc)
            ret._populate_internal_fields()

            # If there is a dclayMethod-decorated __init__, this will call that
            cls.__init__(ret, *args, **kwargs)

            return ret

    def new_dataclay_instance(cls, deserializing, **kwargs):
        """Return a new instance, without calling to the class methods."""
        logger.debug("New dataClay instance (without __call__) of class `%s`", cls.__name__)
        ret = object.__new__(cls)  # this defers the __call__ method
        ret._populate_internal_fields(deserializing=deserializing, **kwargs)
        return ret

    def _prepare_metaclass(cls, namespace, responsible_account):
        """Build a dataClay "MetaClass" for this class.

        :param str namespace: The namespace for this class' MetaClass.
        :param str responsible_account: Responsible account's username.
        :return: A MetaClass Container.
        """
        try:
            class_extradata = cls.get_class_extradata()
        except AttributeError:
            raise ValueError("MetaClass can only be prepared for correctly initialized DataClay Classes")

        logger.verbose("Preparing MetaClass container for class %s (%s)",
                       class_extradata.classname, class_extradata.full_name)

        # The thing we are being asked (this method will keep populating it)
        current_python_info = PythonClassInfo(
            imports=list()
        )
        current_class = MetaClass(
            namespace=namespace,
            name=class_extradata.full_name,
            parentType=None,
            operations=list(),
            properties=list(),
            isAbstract=False,
            languageDepInfos={'LANG_PYTHON': current_python_info}
        )

        ####################################################################
        # OPERATIONS
        ####################################################################
        predicate = inspect.isfunction if six.PY3 else inspect.ismethod
        for name, dataclay_func in inspect.getmembers(cls, predicate):
            # Consider only functions with _dclay_method
            if not getattr(dataclay_func, "_dclay_method", False):
                logger.verbose("Method `%s` doesn't have attribute `_dclay_method`",
                               dataclay_func)
                continue

            original_func = dataclay_func._dclay_entrypoint
            logger.debug("MetaClass container will contain method `%s`, preparing", name)

            # Skeleton for the operation
            current_operation = Operation(
                namespace=namespace,
                className=class_extradata.full_name,
                descriptor=str(),
                signature=str(),
                name=name,
                nameAndDescriptor=name,  # TODO: add descriptor?
                params=dict(),
                paramsOrder=list(),
                returnType=Type.build_from_type(dataclay_func._dclay_ret),
                implementations=list(),
                isAbstract=False,
                isStaticConstructor=False)

            # Start with parameters
            #########################

            # The actual order should be retrieved from the signature
            signature = getfullargspec(original_func)
            if signature.varargs or signature.varkw:
                raise NotImplementedError("No support for varargs or varkw yet")

            current_operation.paramsOrder[:] = signature.args[1:]  # hop over 'self'
            current_operation.params.update({k: Type.build_from_type(v)
                                             for k, v in dataclay_func._dclay_args.items()})

            if len(current_operation.paramsOrder) != len(current_operation.params):
                raise DataClayException("All the arguments are expected to be annotated, " \
                    "there is some error in %s::%s|%s" \
                                        % (namespace, class_extradata.full_name, name))

            # Follow by implementation
            ############################

            current_implementation = PythonImplementation(
                responsibleAccountName=responsible_account,
                namespace=namespace,
                className=class_extradata.full_name,
                opNameAndDescriptor=name,  # TODO: add descriptor?
                position=0,
                includes=list(),
                accessedProperties=list(),
                accessedImplementations=list(),
                requiredQuantitativeFeatures=dict(),
                requiredQualitativeFeatures=dict(),
                code=inspect.getsource(dataclay_func._dclay_entrypoint))

            current_operation.implementations.append(current_implementation)

            # Finally, add the built operation container into the class
            #############################################################
            current_class.operations.append(current_operation)

        ####################################################################
        # PROPERTIES
        ####################################################################
        for n, p in class_extradata.properties.items():
            current_property = Property(
                namespace=namespace,
                className=class_extradata.full_name,
                name=n,
                position=p.position,
                type=p.type,
                beforeUpdate=p.beforeUpdate,
                afterUpdate=p.afterUpdate,
                inMaster=p.inMaster)

            current_class.properties.append(current_property)

        ####################################################################
        # IMPORTS
        ####################################################################
        current_python_info.imports.extend(class_extradata.imports)

        return current_class



"""Initialization and finalization of dataClay client API.

The `init` and `finish` functions are availble through the
dataclay.api package.
"""

from dataclay.commonruntime.Settings import settings
from dataclay.communication.grpc.clients.ExecutionEnvGrpcClient import EEClient
from dataclay.communication.grpc.messages.common.common_messages_pb2 import LANG_PYTHON
from dataclay.serialization.lib.SerializationLibUtils import SerializationLibUtilsSingleton
from dataclay.commonruntime.DataClayRuntime import DataClayRuntime
from dataclay.commonruntime.RuntimeType import RuntimeType
from dataclay.heap.ClientHeapManager import ClientHeapManager
from dataclay.loader.ClientObjectLoader import ClientObjectLoader

# Sentinel-like object to catch some typical mistakes
UNDEFINED_LOCAL = object()


class ClientRuntime(DataClayRuntime):
    
    current_type = RuntimeType.client

    def __init__(self):
        
        DataClayRuntime.__init__(self)
        
    def initialize_runtime_aux(self):
        self.dataclay_heap_manager = ClientHeapManager(self)
        self.dataclay_object_loader = ClientObjectLoader(self)
        
    def is_exec_env(self):
        return False

    def store_object(self, instance):
        raise RuntimeError("StoreObject can only be used from the ExecutionEnvironment")
    
    def choose_location(self, instance, alias=None):
        """ Choose execution/make persistent location. 
        :param instance: Instance to use in call
        :param instance: The alias of the instance
        :returns: Location 
        :type instance: DataClayObject
        :rtype: DataClayID 
        """
        
        if alias:
            instance._update_object_id(self.get_object_id_by_alias(alias))
        
        exec_env_id = self.get_object_location_by_id(instance.get_object_id())
        instance.set_hint(exec_env_id)
        self.logger.verbose("ExecutionEnvironmentID obtained for execution = %s", exec_env_id)
        return exec_env_id
    
    def make_persistent(self, instance, alias, backend_id, recursive):
        """ This method creates a new Persistent Object using the provided stub
        instance and, if indicated, all its associated objects also Logic module API used for communication
        This function is called from a stub/execution class
        :param instance: Instance to make persistent
        :param backend_id: Indicates which is the destination backend
        :param recursive: Indicates if make persistent is recursive
        :param alias: Alias for the object
        :returns: ID of the backend in which te object was persisted.
        :type instance: DataClayObject
        :type backend_id: DataClayID
        :type recursive: boolean
        :type alias: string
        :rtype: DataClayID
        :raises RuntimeError: if backend id is UNDEFINED_LOCAL.
        """
        self.logger.debug("Starting make persistent object for instance %s with id %s", instance,
                     instance.get_object_id())
        if backend_id is UNDEFINED_LOCAL:
            # This is a commonruntime end user pitfall,
            # @abarcelo thinks that it is nice
            # (and exceptionally detailed) error
            raise RuntimeError("""
                You are trying to use dataclay.api.LOCAL but either:
                  - dataClay has not been initialized properly
                  - LOCAL has been wrongly imported.
                
                Be sure to use LOCAL with:
                
                from dataclay import api
                
                and reference it with `api.LOCAL`
                
                Refusing the temptation to guess.""")
        location = instance.get_hint()
        if location is None:
            location = backend_id
            # Choose location if needed
            # If object is already persistent -> it must have a Hint (location = hint here)
            # If object is not persistent -> location is choosen (provided backend id or random, hash...).
            if location is None:
                location = self.choose_location(instance, alias)
            
        if not instance.is_persistent():

            # === MAKE PERSISTENT === #
            
            # set the default master location
            instance.set_master_location(location)
            # We serialize objects like volatile parameters
            parameters = list()
            parameters.append(instance)
            params_order = list()
            params_order.append("object")
            params_spec = dict()
            params_spec["object"] = "DataClayObject"  # not used, see serialized_params_or_return
            serialized_objs = SerializationLibUtilsSingleton.serialize_params_or_return(
                params=parameters,
                iface_bitmaps=None,
                params_spec=params_spec,
                params_order=params_order,
                hint_volatiles=location,
                runtime=self,
                recursive=recursive)
            
            # Avoid some race-conditions in communication (make persistent + execute where
            # execute arrives before).
            self.add_volatiles_under_deserialization(serialized_objs[3])
            
            # Get EE
            try:
                execution_client = self.ready_clients[location]
            except KeyError:
                exeenv = self.get_execution_environments_info()[location] 
                self.logger.debug("Not found in cache ExecutionEnvironment {%s}! Starting it at %s:%d",
                                   location, exeenv.hostname, exeenv.port)
                execution_client = EEClient(exeenv.hostname, exeenv.port)
                self.ready_clients[location] = execution_client
    
            # Call EE
            self.logger.verbose("Calling make persistent to EE %s ", location)
            execution_client.make_persistent(settings.current_session_id, serialized_objs)

            # update the hint with the location, and return it
            instance.set_hint(location)
            
            # remove volatiles under deserialization
            self.remove_volatiles_under_deserialization()
            
        if alias is not None:
            # Add a new alias to an object.
            # Use cases:
            # 1 - object was persisted without alias and not yet registered -> we need to register it with new alias.
            # 2 - object was persisted and it is already registered -> we only add a new alias
            # 3 - object was persisted with an alias and it must be already registered -> we add a new alias.

            # From client side, we cannot check if object is registered or not (we do not have isPendingToRegister like EE)
            # Therefore, we call LogicModule with all information for registration.
            reg_info = [instance.get_object_id(), instance.get_class_extradata().class_id,
                        self.get_session_id(), instance.get_dataset_id()]
            self.ready_clients["@LM"].register_object(reg_info, location, alias, LANG_PYTHON)
        
            self.alias_cache[alias] = instance.get_object_id(), instance.get_class_extradata().class_id, location
        
        return location
    
    def add_session_reference(self, object_id):
        """
        @summary Add +1 reference associated to thread session
        @param object_id ID of object.
        """
        pass
    
    def execute_implementation_aux(self, operation_name, instance, parameters, exeenv_id=None):
        stub_info = instance.get_class_extradata().stub_info
        implementation_stub_infos = stub_info.implementations
        object_id = instance.get_object_id()
        
        self.logger.verbose("Calling operation '%s' in object with ID %s", operation_name, object_id)
        self.logger.debug("Call is being done into %r with #%d parameters",
                          instance, len(parameters))
        
        using_hint = False
        if instance.get_hint() is not None:
            exeenv_id = instance.get_hint()
            self.logger.verbose("Using hint = %s", exeenv_id)
        else:
            exeenv_id = next(iter(self.get_metadata(object_id).locations))
            
        return self.call_execute_to_ds(instance, parameters, operation_name, exeenv_id, using_hint)

    def get_operation_info(self, object_id, operation_name):
        dcc_extradata = self.get_object_by_id(object_id).get_class_extradata()
        stub_info = dcc_extradata.stub_info
        implementation_stub_infos = stub_info.implementations
        operation = implementation_stub_infos[operation_name]
        return operation 
    
    def get_implementation_id(self, object_id, operation_name, implementation_idx=0):
        operation = self.get_operation_info(object_id, operation_name)
        return operation.remoteImplID
    
    def close_session(self):
         self.ready_clients["@LM"].close_session(settings.current_session_id)

    def get_hint(self):
        return None
    
    def get_session_id(self):
        return settings.current_session_id
    




""" Class description goes here. """

"""Management of dataClay MetaClass containers.

Note that this is NOT for Python "metaclass", but for the MetaClass container
which includes information of registered classes.

Take into account that managing MetaClass container tends to be a job for the
ExecutionEnvironment, not for the client. But with notable exceptions (like the
class registration, however it doesn't use those functions at the moment).

It is in the library instead of inside executionenv because the ExecutionGateway
requires some specific methods from here. But this may change in the future.
"""

from dataclay.commonruntime.Runtime import getRuntime
from dataclay.commonruntime.Settings import settings
from dataclay.util.YamlParser import dataclay_yaml_load
from lru import LRU
import os.path
import six

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

# TODO: un-hardcode this
cached_metaclasses = LRU(200)
cached_metaclass_info = LRU(200)

        
def deploy_metaclass_grpc(namespace, class_name, metaclass_yaml_str, metaclass):
    """Deploy binary data into the internal structure (currently, a file)
    :param namespace: The name of the namespace for the class.
    :param class_name: Name of the class.
    :param metaclass: a MetaClass as YAML string.

    Note that there is redundancy on the parameters because this is typically
    called after a server-side call to deployment, and thus the info is
    available in multiple form.
    """
    namespace_path = os.path.join(settings.deploy_path, namespace)
    if not os.access(namespace_path, os.X_OK | os.W_OK):
        # Assume that it is not created yet
        os.makedirs(namespace_path)
        if not os.access(namespace_path, os.X_OK | os.W_OK):
            raise OSError("Could not create/have write access to folder {}".
                          format(namespace_path))

    # Store in a file
    with open(os.path.join(namespace_path, class_name + ".mcs"), 'wb') as f:
        if six.PY2:
            f.write(bytes(metaclass_yaml_str))
        elif six.PY3:    
            f.write(bytes(metaclass_yaml_str, 'utf-8'))

    if settings.cache_on_deploy:
        cached_metaclasses[class_name] = metaclass


def load_metaclass(namespace, class_name):
    """Load binary data into a MetaClass container.
    :param namespace: The name of the namespace for the class.
    :param class_name: Name of the class.
    :return: A MetaClass container, deserialized from previously deployed binary data.
    """
    try:
        return cached_metaclasses[class_name]
    except KeyError:
        namespace_path = os.path.join(settings.deploy_path, namespace)
        with open(os.path.join(namespace_path, class_name + ".mcs"), 'rb') as f:
            mc = dataclay_yaml_load(f.read())
        cached_metaclasses[class_name] = mc
        return mc


def load_metaclass_info(metaclass_id):
    """Load the namespace and class name for a certain MetaClassID.
    :param metaclass_id: The dataClay UUID of the MetaClass.
    :return: A tuple (class_name, namespace).
    """
    try:
        return cached_metaclass_info[metaclass_id]
    except KeyError:
        class_name, namespace = getRuntime().ready_clients["@LM"].get_classname_and_namespace_for_ds(metaclass_id)
        cached_metaclass_info[metaclass_id] = (class_name, namespace)
        return class_name, namespace



"""Submodule for all the YAML-based management structures.

For specific information about the structures, see original Java implementation
--it is used as the authoritative implementation, available in util.management.

Note that the YAML dictionary-based fields are listed in the _fields attribute
and this is what is used on the load & dump YAML procedures.

For more information of the internal magic, see "MgrObject.py" file.

BLACK MAGIC WARNING! The NAME of the module (like "contractmgr") and the NAME
of the class (like "Contract") are used in order to obtain the YAML tag, which
will (hopefully) match the Java generated name, which comes from the Java
package name (like "!!util.management.contractmgr.Contract"). This is syntactic
sugar, but should be taken into account for cross-language correctness.
"""
from collections import namedtuple
from functools import partial
from yaml import Loader, load, dump

from dataclay.util.YamlIgnores import IGNORE_CLASSES, IGNORE_PREFIXES

from .ids import _uuid
from .info.VersionInfo import VersionInfo
from .management.accountmgr.Account import Account
from .management.classmgr.AccessedImplementation import AccessedImplementation
from .management.classmgr.AccessedProperty import AccessedProperty
from .management.classmgr.Implementation import Implementation
from .management.classmgr.MetaClass import MetaClass
from .management.classmgr.Operation import Operation
from .management.classmgr.Property import Property
from .management.classmgr.Type import Type
from .management.classmgr.UserType import UserType
from .management.classmgr.java.JavaImplementation import JavaImplementation
from .management.classmgr.python.PythonClassInfo import PythonClassInfo
from .management.classmgr.python.PythonImplementation import PythonImplementation
from .management.contractmgr.Contract import Contract
from .management.datacontractmgr.DataContract import DataContract
from .management.datasetmgr.DataSet import DataSet
from .management.interfacemgr.Interface import Interface
from .management.metadataservice.ExecutionEnvironment import ExecutionEnvironment
from .management.metadataservice.MetaDataInfo import MetaDataInfo
from .management.metadataservice.StorageLocation import StorageLocation
from .management.metadataservice.DataClayInstance import DataClayInstance
from .management.namespacemgr.Namespace import Namespace
from .management.sessionmgr.SessionContract import SessionContract
from .management.sessionmgr.SessionDataContract import SessionDataContract
from .management.sessionmgr.SessionImplementation import SessionImplementation
from .management.sessionmgr.SessionInfo import SessionInfo
from .management.sessionmgr.SessionInterface import SessionInterface
from .management.sessionmgr.SessionOperation import SessionOperation
from .management.sessionmgr.SessionProperty import SessionProperty
from .management.stubs.ImplementationStubInfo import ImplementationStubInfo
from .management.stubs.PropertyStubInfo import PropertyStubInfo
from .management.stubs.StubInfo import StubInfo
from dataclay.communication.grpc.messages.common.common_messages_pb2 import Langs


def trivial_constructor(loader, node):
    """Constructor used to "ignore" certain types.

    The behaviour is always to build a mapping. This is a harmless behaviour
    (at least, is expected to be). dataClay uses this for all want-to-ignore
    types, without losing semantics. If problems arise from this, this method
    could avoid them by returning None.

    For aesthetic reasons, a namedtuple instance is returned (which will be
    built tailored to the input, which may or may not be an expected behaviour)
    and its name will be used from the tag.
    """
    name = node.tag.rsplit(".", 1)[-1]
    contents = loader.construct_mapping(node)
    return namedtuple(name, contents.keys())(**contents)


def tuple_constructor(loader, node):
    ''' Constructor for a Java Tuple represented in YAML, which is is a simple two-element python tuple.
    :param loader: 
    :param node:
    :returns: None
    :rtype: None
    '''
    d = loader.construct_mapping(node)
    return d["first"], d["second"]


def feature_constructor(loader, node):
    """Feature (Java enum) is parsed as a String."""
    s = loader.construct_scalar(node)
    # Ugly hack, but we do not need the value at all
    return hash(s)


def lonely_equal_constructor(loader, node):
    """Solve/monkey-patch a very old bug.
    https://bitbucket.org/xi/pyyaml/issues/49/plain-equal-sign-as-node-content-results
    """
    s = loader.construct_scalar(node)
    return s


def lang_constructor(loader, node):
    """Language is parsed as a GRPC enum."""
    s = loader.construct_scalar(node)
    return Langs.Value(s)


Loader.add_constructor(u"tag:yaml.org,2002:value", lonely_equal_constructor)

# The tuple is a bit special itself
Loader.add_constructor(u"tag:yaml.org,2002:es.bsc.dataclay.util.structs.Tuple", tuple_constructor)

# Not needed for Python, but nice to avoid errors
Loader.add_constructor(u"tag:yaml.org,2002:es.bsc.dataclay.util.management.classmgr.features.Feature$FeatureType",
                       feature_constructor)

# The language is very special
Loader.add_constructor(u"tag:yaml.org,2002:es.bsc.dataclay.communication.grpc.messages.common.CommonMessages$Langs", lang_constructor)

for prefix in IGNORE_PREFIXES:
    yaml_tag_prefix = u"tag:yaml.org,2002:%s" % prefix
    Loader.add_multi_constructor(
        yaml_tag_prefix,
        # The following is used to disregard the tag
        lambda loader, _, node: trivial_constructor(loader, node))

for class_tag in IGNORE_CLASSES:
    yaml_class_tag = u"tag:yaml.org,2002:%s" % class_tag
    Loader.add_constructor(yaml_class_tag, trivial_constructor)
    
# Force all YAML usages to go through this class (so Loader is properly set)
dataclay_yaml_load = partial(load, Loader=Loader)
dataclay_yaml_dump = dump



from dataclay import DataClayObject, dclayMethod


class StorageList(DataClayObject):
    """Demonstration of the StorageList class.

    This class is prepared for demonstration purposes and is not suitable,
    as it is right now, for HPC applications and production workflows.

    @ClassField _list anything
    """

    @dclayMethod()
    def __init__(self):
        self._list = list()

    @dclayMethod(return_="int")
    def __len__(self):
        return len(self._list)
    
    @dclayMethod(item="anything", return_="bool")
    def __contains__(self, item):
        return item in self._list

    @dclayMethod(item="anything")
    def append(self, item):
        self._list.append(item)

    @dclayMethod(_local=True, return_="anything")
    def __iter__(self):
        return iter(self._list)

    @dclayMethod(_local=True, return_="anything")
    def split(self):
        # Ugly split-in-two, for demonstration purposes only
        from itertools import cycle

        out_a = list()
        out_b = list()

        for elem, out in zip(self._list, cycle(out_a, out_b)):
            out.append(elem)

        return (out_a, out_b)

    @dclayMethod(item="anything", return_="anything")
    def __getitem__(self, item):
        return self._list[item]

    @dclayMethod(item="anything", value="anything")
    def __setitem__(self, item, value):
        self._list[item] = value

    @dclayMethod(item="anything")
    def __delitem__(self, item):
        del self._list[item]

    @dclayMethod(return_="str")
    def __str__(self):
        return str(self._list)



""" Class description goes here. """

from dataclay.serialization.python.DataClayPythonWrapper import DataClayPythonWrapper

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'


class NotImplementedType(DataClayPythonWrapper):
    __slots__ = ()

    def __init__(self):
        pass

    def read(self, io_file):
        raise NotImplementedError("Trying to deserialize something not implemented")

    def write(self, io_file, value):
        raise NotImplementedError("Trying to serialize something not implemented")



""" Class description goes here. """

import dataclay.api
from distutils.util import strtobool
import os
import logging
import uuid

from dataclay.commonruntime.Runtime import getRuntime

# "Publish" the StorageObject (which is a plain DataClayObject internally)
from dataclay import DataClayObject as StorageObject
# Also "publish" the split method
from dataclay.contrib.splitting import split

# The StorageDict and StorageList data structures
from .models.storagedict import StorageDict
from .models.storagelist import StorageList

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

_initialized = False

logger = logging.getLogger('dataclay.storage.api')


def getByID(object_strid):
    """Get a Persistent Object from its OID.
    :param object_strid: The string identifying object (contains both ObjectID and hint)
    :return: The (Persistent) DataClayObject
    """
    try:
        object_id, hint, class_id = object_strid.split(":")
        ret = getRuntime().get_object_by_id(uuid.UUID(object_id),
                                            class_id=uuid.UUID(class_id),
                                            hint=uuid.UUID(hint))
    except ValueError:  # this can fail for both [not enough semicolons]|[invalid uuid]
        # Fallback behaviour: no extra fields, the whole string is the ObjectID UUID
        object_id = object_strid
        ret = getRuntime().get_object_by_id(uuid.UUID(object_id))

    return ret


def initWorker(config_file_path, **kwargs):
    """Worker-side initialization.

    CURRENT BEHAVIOUR IMPORTANT CONSIDERATIONS

    This initialization is **not** a complete one, this performs a
    `per_network_init`ialization. This is done because it is assumed that
    COMPSs will proceed to fork the process and then call the
    `initWorkerPostFort`.

    Be sure to leave all the process fragile stuff in the other call. But keep
    in mind that calling to initWorkerPostFor is not optional --after this init
    the library is not in a consistent state.

    :param config_file_path: Path to storage.properties configuration file.
    :param kwargs: Additional arguments, currently unused. For future use
      and/or other Persistent Object Library requirements.
    """
    logger.info("Initialization of worker through storage.api")
    dataclay.api.pre_network_init(config_file_path)


def initWorkerPostFork():
    """Worker-side initialization after forking the process.

    CURRENT BEHAVIOUR IMPORTANT CONSIDERATIONS

    This initialization must be done after the `initWorker` and all the
    libraries that are "fork-fragile" must be addressed here.

    To this date, only gRPC client is initialized here because they are not
    process-safe.

    Once this method is called, dataClay can be considered "initialized".
    """
    logger.warning("Finishing initialization (post-fork)")
    dataclay.api.reinitialize_logging()

    # ToDo: this is ugly, avoid, fix the issue:
    # https://storage.bsc.es/gitlab/object_storage/poas/issues/78
    logger.warning("Arbitrary sleep done here in order to minimize "
                   "race conditions on opened connections to management database")
    from time import sleep
    from random import randint
    sleep(randint(0, 30))
    ##########################################################
    dataclay.api.post_network_init()
    logger.debug("Reinitialization post-fork finished")


def finishWorkerPostFork():
    """Worker-side finalization per forked process.

    This finalization must be done before the finishWorker. This must be
    called once per forked process, and must match the calls of
    initWorkerPostFork.

    That means that each call to initWorkerPostFork must have its "related"
    (i.e. called from the same process) call to finishWorkerPostFork.
    """
    logger.warning("Finishing dataClay (post-fork)")
    dataclay.api.finish()
    logger.debug("Finalization post-fork finished")


def init(config_file_path, **kwargs):
    """Master-side initialization.

    Identical to initWorker (right now, may change).
    """
    logger.info("Initialization of storage.api")
    dataclay.api.init(config_file_path)


def finishWorker(**kwargs):
    """Worker-side finalization.

    :param kwargs: Additional arguments, currently unused. For future use
      and/or other Persistent Object Library requirements.
    """
    logger.info("Finalization of worker through storage.api")
    # Nothing to do here, because finishWorkerPostFork is the real hero here
    dataclay.api.finish_tracing()  # Not enough hero for Extrae :)


def finish(**kwargs):
    """Master-side initialization.

    Identical to finishworker (right now, may change).
    """
    logger.info("Finalization of storage.api")
    dataclay.api.finish()


class TaskContext(object):

    def __init__(self, logger, values, config_file_path=None, **kwargs):
        """Initialize the TaskContext for the current task.
        :param logger: A logger that can be used for the storage.api
        :param values: The values of the Task (unused right now)
        :param config_file_path: DEPRECATED. Was required by dataClay.
        Has been moved to initWorker.
        :param kwargs: Additional arguments, currently unused. For future use
        and/or other Persistent Object Library requirements.

        A task is about to be executed and this ContextManager encloses its
        execution. This context is used by the COMPSs Worker.
        """
        self.logger = logger
        self.values = values

    def __enter__(self):
        """Perform initialization (ContextManager starts)"""
        self.logger.info("Starting task")
        logger.info("Starting task")

    def __exit__(self, etype, value, tb):
        """Perform finalization (ContextManager ends)"""
        # ... manage exception if desired
        if etype is not None:
            self.logger.warn("Exception received: %s", etype)
            logger.warn("Exception received: %s", etype)

            import traceback
            traceback.print_exception(etype, value, tb)

            pass  # Exception occurred
            # Return true if you want to suppress the exception

        # Finished
        self.logger.info("Ending task")
        logger.info("Ending task")


if strtobool(os.getenv("DEACTIVATE_STORAGE_LIBRARY", "False")):
    from dataclay.contrib.dataclay_dummy import deactivate_storage_library
    deactivate_storage_library()




""" Class description goes here. """

"""Management of Persistent Objects."""

import collections

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2016 Barcelona Supercomputing Center (BSC-CNS)'


class IdentityDict(collections.MutableMapping):
    __slots__ = ("internal_dict",)

    def __init__(self):
        self.internal_dict = dict()

    def __setitem__(self, key, value):
        self.internal_dict[id(key)] = (key, value)

    def __getitem__(self, item):
        return self.internal_dict[id(item)][1]

    def items(self):
        return self.internal_dict.values()

    def __delitem__(self, key):
        del self.internal_dict[id(key)]

    def __len__(self):
        return len(self.internal_dict)

    # ToDo: this is not a valid canonical iteration of a dictionary (should be only the key)
    def __iter__(self):
        return self.internal_dict.values()

    def __contains__(self, item):
        return id(item) in self.internal_dict




""" Class description goes here. """

"""Automatic population helpers for the settings.

When certain unknown setting is asked (like account UUID), the classes defined
here help to automatically populate them.
"""
from abc import ABCMeta, abstractmethod

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'
from dataclay.commonruntime.Runtime import getRuntime
import six

@six.add_metaclass(ABCMeta)
class AbstractLoader(object):
    """An abstraction to allow lazy evaluation of settings fields.

    See dataclay.conf._SettingsHub to check the behaviour and purpose of this
    class. The different loaders expose an abstraction to allow lazy and late
    initialization of variable values.
    """

    @abstractmethod
    def __init__(self, settings_object):
        self._settings = settings_object

    @abstractmethod
    def load_value(self):
        """Perform the active load of the requested value.

        This is typically executed only once, as the settings will store the
        real object in its place.

        :return: The expected commonruntime value for the settings field.
        """
        pass


class AccountIdLoader(AbstractLoader):

    def __init__(self, settings_object, field_of_account_name):
        """Load an Account ID for a certain user

        :param field_of_account_name: The name of the field of the settings
        that contain the account name.
        """
        self._field = field_of_account_name
        super(AccountIdLoader, self).__init__(settings_object)

    def load_value(self):
        account_name = getattr(self._settings, self._field)
        return getRuntime().ready_clients["@LM"].get_account_id(account_name)


class AccountCredentialLoader(AbstractLoader):

    def __init__(self, settings_object, field_of_account_id, field_of_account_password):
        """Prepare a credential pair for a certain account.
        :param field_of_account_id: The name of the field of the settings
        that contain the account id.
        :param field_of_account_password: Same but for the password.
        that contain the account name.
        """
        self._field_id = field_of_account_id
        self._field_password = field_of_account_password
        super(AccountCredentialLoader, self).__init__(settings_object)

    def load_value(self):
        return (getattr(self._settings, self._field_id),
                getattr(self._settings, self._field_password))
